import pytest

from leapp.libraries.actor import rocescanner
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked, produce_mocked
from leapp.libraries.stdlib import CalledProcessError

NMCLI_CON_NIC1 = [
    'ens1: connected to ens1',
    '"Mellanox MT27710"',
    'ethernet (mlx5_core), 82:28:9B:1B:28:2C, hw, mtu 1500',
    'inet4 192.168.0.2/24',
    'route4 192.168.0.1/24',
    'inet6 fe80::d8c5:3a67:1abb:dcca/64',
    'route6 fe80::/64',
    ''
]

NMCLI_CON_NIC2 = [
    'eno2: connected to eno2',
    '"mellanox MT27710"',
    'ethernet (mlx5_core), 82:28:9B:1B:28:2C, hw, mtu 1500',
    'inet4 172.18.0.2/16',
    'route4 172.18.0.1/16',
    'inet6 fe80::d8c5:3a67:1abb:dcca/64',
    'route6 fe80::/64',
    ''
]

NMCLI_DISCON_NIC3 = [
    'ens3: disconnected',
    '"Mellanox MT27710"',
    'ethernet (mlx5_core), 82:28:9B:1B:28:2C, hw, mtu 1500',
    ''
]

NMCLI_CON_NIC4 = [
    'mellanox4: connecting',
    '"Mellanox MT27710"',
    'ethernet (mlx5_core), 82:28:9B:1B:28:2C, hw, mtu 1500',
    'inet4 192.168.0.1/16',
    'route4 192.168.0.1/16',
    'inet6 fe80::d8c5:3a67:1abb:dcca/64',
    'route6 fe80::/64',
    ''
]

NMCLI_CON_NIC5_NO_ROCE = [
    'ens5: connected to ens5',
    '"Red Hat Virtio"',
    'ethernet (mlx5_core), 82:28:9B:1B:28:2C, hw, mtu 1500',
    'inet4 192.168.0.1/16',
    'route4 192.168.0.1/16',
    'inet6 fe80::d8c5:3a67:1abb:dcca/64',
    'route6 fe80::/64',
    ''
]


@pytest.mark.parametrize('nmcli_stdout,expected', (
    ([], []),
    (NMCLI_CON_NIC5_NO_ROCE, []),
    # simple
    (NMCLI_CON_NIC1, [NMCLI_CON_NIC1[0]]),
    (NMCLI_CON_NIC2, [NMCLI_CON_NIC2[0]]),
    (NMCLI_CON_NIC4, [NMCLI_CON_NIC4[0]]),
    (NMCLI_DISCON_NIC3, [NMCLI_DISCON_NIC3[0]]),
    # multiple
    (
        NMCLI_CON_NIC1 + NMCLI_CON_NIC2,
        [NMCLI_CON_NIC1[0], NMCLI_CON_NIC2[0]]
    ),
    (
        NMCLI_CON_NIC1 + NMCLI_DISCON_NIC3,
        [NMCLI_CON_NIC1[0], NMCLI_DISCON_NIC3[0]]
    ),
    (
        NMCLI_CON_NIC5_NO_ROCE + NMCLI_CON_NIC2,
        [NMCLI_CON_NIC2[0]]
    ),
    (
        NMCLI_CON_NIC2 + NMCLI_CON_NIC5_NO_ROCE,
        [NMCLI_CON_NIC2[0]]
    ),
))
def test_get_roce_nics_lines(monkeypatch, nmcli_stdout, expected):
    def mocked_run(cmd, *args, **kwargs):
        assert cmd == ['nmcli']
        return {'stdout': nmcli_stdout}
    monkeypatch.setattr(rocescanner, 'run', mocked_run)
    assert rocescanner.get_roce_nics_lines() == expected


@pytest.mark.parametrize('raise_exc', (
    CalledProcessError('foo', {'stdout': '', 'stderr': 'err', 'exit_code': '1'}, ['nmcli']),
    OSError('foo')
))
def test_get_roce_nics_lines_err(monkeypatch, raise_exc):
    def mocked_run(cmd, *args, **kwargs):
        assert cmd == ['nmcli']
        raise raise_exc
    monkeypatch.setattr(rocescanner, 'run', mocked_run)
    monkeypatch.setattr(rocescanner.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(rocescanner.api, 'current_actor', CurrentActorMocked())
    assert rocescanner.get_roce_nics_lines() == []
    assert rocescanner.api.current_logger.warnmsg


@pytest.mark.parametrize('roce_lines,connected,connecting', (
    ([], [], []),
    ([NMCLI_DISCON_NIC3[0]], [], []),
    ([NMCLI_CON_NIC1[0]], ['ens1'], []),
    ([NMCLI_CON_NIC2[0]], ['eno2'], []),
    ([NMCLI_CON_NIC4[0]], [], ['mellanox4']),
    (
        [
            'ens1: connected to ens1',
            'eno2: connecting',
            'route6 fe80::/64',
            '',
            'ens3: connected to ens3',
        ],
        ['ens1', 'ens3'],
        ['eno2']
    ),
))
def test_roce_detected(monkeypatch, roce_lines, connected, connecting):
    mocked_produce = produce_mocked()
    monkeypatch.setattr(rocescanner.api, 'current_actor', CurrentActorMocked(arch=architecture.ARCH_S390X))
    monkeypatch.setattr(rocescanner.api.current_actor(), 'produce', mocked_produce)
    monkeypatch.setattr(rocescanner, 'get_roce_nics_lines', lambda: roce_lines)
    rocescanner.process()
    if connected or connecting:
        assert mocked_produce.called
        msg = mocked_produce.model_instances[0]
        assert msg.roce_nics_connected == connected
        assert msg.roce_nics_connecting == connecting
    else:
        assert not mocked_produce.called


@pytest.mark.parametrize('arch', (
    architecture.ARCH_ARM64,
    architecture.ARCH_X86_64,
    architecture.ARCH_PPC64LE
))
def test_roce_noibmz(monkeypatch, arch):
    def mocked_roce_lines():
        assert False, 'Unexpected call of get_roce_nics_lines on nonIBMz arch.'
    mocked_produce = produce_mocked()
    monkeypatch.setattr(rocescanner.api, 'current_actor', CurrentActorMocked(arch=arch))
    monkeypatch.setattr(rocescanner.api.current_actor(), 'produce', mocked_produce)
    monkeypatch.setattr(rocescanner, 'get_roce_nics_lines', lambda: mocked_roce_lines)
    assert not mocked_produce.called



from leapp.actors import Actor
from leapp.libraries.stdlib import CalledProcessError, run
from leapp.models import NetworkManagerConfig
from leapp.tags import FirstBootPhaseTag, IPUWorkflowTag


class NetworkManagerUpdateConnections(Actor):
    """
    Update NetworkManager connections.

    When using dhcp=dhclient on Red Hat Enterprise Linux 7, a non-hexadecimal client-id (a string)
    is sent on the wire as is (i.e. the first character is the 'type' as per RFC 2132 section
    9.14). On Red Hat Enterprise Linux 8, a zero byte is prepended to string-only client-ids. To
    preserve behavior on upgrade, we convert client-ids to the hexadecimal form.
    """

    name = 'network_manager_update_connections'
    consumes = (NetworkManagerConfig,)
    produces = ()
    tags = (FirstBootPhaseTag, IPUWorkflowTag)

    def process(self):
        for nm_config in self.consume(NetworkManagerConfig):
            if nm_config.dhcp not in ('', 'dhclient'):
                self.log.info('DHCP client is "{}", nothing to do'.format(nm_config.dhcp))
                return

            try:
                r = run(['/usr/bin/python3', 'tools/nm-update-client-ids.py'])['stdout']
                self.log.info('Updated client-ids: {}'.format(r))
            except (OSError, CalledProcessError) as e:
                self.log.warning('Error calling nm-update-client-ids script: {}'.format(e))

            break



import os
import re
import shutil

from leapp.libraries.common.config import version
from leapp.libraries.stdlib import api, CalledProcessError, run

DAEMON_FILE = '/etc/frr/daemons'
# if this file still exists after the removal of quagga, it has been modified
CONFIG_FILE = '/etc/sysconfig/quagga.rpmsave'
QUAGGA_CONF_FILES = '/etc/quagga/'
FRR_CONF_FILES = '/etc/frr/'
BGPD_CONF_FILE = '/etc/frr/bgpd.conf'

regex = re.compile(r'\w+(?<!WATCH)(?<!BABELD)_OPTS=".*"')


def _get_config_data(path):
    conf_data = {}
    with open(path) as f:
        for line in f:
            if regex.match(line):
                k, v = line.rstrip().split("=")
                conf_data[k.split("_")[0].lower()] = v.strip('"')

    return conf_data


def _edit_new_config(path, active_daemons, config_data):
    with open(path, 'r') as f:
        data = f.read()

    # replace no as yes in /etc/frr/daemons
    for daemon in active_daemons:
        data = re.sub(r'{}=no'.format(daemon), r'{}=yes'.format(daemon), data, flags=re.MULTILINE)

    if config_data:
        for daemon in config_data:
            data = re.sub(r'{}_options=\(".*"\)'.format(daemon),
                          r'{}_options=("{}")'.format(daemon, config_data[daemon]),
                          data, flags=re.MULTILINE)

    return data


# 1. parse /etc/sysconfig/quagga.rpmsave if it exists
# 2. change =no to =yes with every enabled daemon
# 3. use data from data from quagga.rpmsave in new daemon file
def _change_config(quagga_facts):
    config_data = {}
    if os.path.isfile(CONFIG_FILE):
        config_data = _get_config_data(CONFIG_FILE)

    # This file should definitely exist, if not, something went wrong with the upgrade
    if os.path.isfile(DAEMON_FILE):
        data = _edit_new_config(DAEMON_FILE, quagga_facts.active_daemons, config_data)
        with open(DAEMON_FILE, 'w') as f:
            f.write(data)


# In quagga, each daemon needed to be started individually
# In frr, only frr is started as a daemon, the rest is started based on the daemons file
# So as long as at least one daemon was active in quagga, frr should be enabled
def _enable_frr(quagga_facts):
    # remove babeld?
    if quagga_facts.enabled_daemons:
        try:
            run(['systemctl', 'enable', 'frr'])
        except CalledProcessError:
            return False

    return True


# due to an error in quagga, the conf files are not deleted after uninstall
# we can copy them as they are
def _copy_config_files(src_path, dest_path):
    conf_files = os.listdir(src_path)
    for file_name in conf_files:
        full_path = os.path.join(src_path, file_name)
        if os.path.isfile(full_path):
            shutil.copy(full_path, dest_path)
            api.current_logger().debug('Copying %s to %s%s', full_path, dest_path, file_name)


# some commands in *.conf files have changed in the latest rebase
def _fix_commands():
    if version.matches_target_version(">= 8.4"):
        if os.path.isfile(BGPD_CONF_FILE):
            with open(BGPD_CONF_FILE, 'r') as f:
                data = f.read()
                data = re.sub("ip extcommunity-list", "bgp extcommunity-list", data, flags=re.MULTILINE)
            with open(BGPD_CONF_FILE, 'w') as f:
                f.write(data)


def process_facts(quagga_facts):
    _change_config(quagga_facts)
    _copy_config_files(QUAGGA_CONF_FILES, FRR_CONF_FILES)
    _fix_commands()
    _enable_frr(quagga_facts)



import leapp.libraries.common.multipathutil as lib


def test_no_config():
    content = lib.read_config('/this/does/not/exist')
    assert content is None


def test_blank_line():
    data = lib.LineData('', None, False)
    assert data.type == data.TYPE_BLANK
    data = lib.LineData('    ', 'devices', True)
    assert data.type == data.TYPE_BLANK
    data = lib.LineData('# comment at start of line', 'defaults', False)
    assert data.type == data.TYPE_BLANK
    data = lib.LineData('  ! alternate comment method!!!', 'blacklist', True)
    assert data.type == data.TYPE_BLANK
    data = lib.LineData('# } still a comment', None, False)
    assert data.type == data.TYPE_BLANK
    data = lib.LineData('   # "also a comment!"', "multipaths", True)
    assert data.type == data.TYPE_BLANK
    data = lib.LineData(' "ignore lines starting with a string"', None, False)
    assert data.type == data.TYPE_BLANK


def test_section_end():
    data = lib.LineData('}', None, False)
    assert data.type == data.TYPE_SECTION_END
    data = lib.LineData('     }  # trailing comment', 'devices', True)
    assert data.type == data.TYPE_SECTION_END
    data = lib.LineData('     }  JUNK AFTERWARDS', 'defaults', False)
    assert data.type == data.TYPE_SECTION_END
    data = lib.LineData('\t \t}  !!tabs', 'multipaths', True)
    assert data.type == data.TYPE_SECTION_END


def test_section_start():
    sections = ('defaults', 'blacklist', 'blacklist_exceptions', 'devices',
                'overrides', 'multipaths')
    subsections = {'blacklist': 'device', 'blacklist_exceptions': 'device',
                   'devices': 'device', 'multipaths': 'multipath'}
    for section in sections:
        data = lib.LineData(section + ' {', None, False)
        assert data.type == data.TYPE_SECTION_START
        assert data.section == section

    try:
        data = lib.LineData('not_a_section {', None, False)
        assert False
    except ValueError:
        pass

    data = lib.LineData('defaults # works even without brace', None, False)
    assert data.type == data.TYPE_SECTION_START
    assert data.section == 'defaults'

    data = lib.LineData('\t \tmultipaths { # tabs', None, False)
    assert data.type == data.TYPE_SECTION_START
    assert data.section == 'multipaths'

    data = lib.LineData('devices{ # do not need a space', None, False)
    assert data.type == data.TYPE_SECTION_START
    assert data.section == 'devices'

    for section in subsections:
        data = lib.LineData(subsections[section] + ' {', section, False)
        assert data.type == data.TYPE_SECTION_START
        assert data.section == subsections[section]

    data = lib.LineData('devices { # wrong section', 'multipath', False)
    assert data.type != data.TYPE_SECTION_START

    data = lib.LineData('devices { # already in subsection', 'device', True)
    assert data.type != data.TYPE_SECTION_START


def test_option():
    data = lib.LineData('key value', 'defaults', False)
    assert data.type == data.TYPE_OPTION
    assert data.option == 'key'
    assert data.value == 'value'

    data = lib.LineData('key value # comment', 'devices', True)
    assert data.option == 'key'
    assert data.value == 'value'

    data = lib.LineData('    key value "extra string"', 'multipaths', True)
    assert data.option == 'key'
    assert data.value == 'value'

    data = lib.LineData(' \t\tkey "string value" junk', 'blacklist', False)
    assert data.option == 'key'
    assert data.value == 'string value'

    try:
        data = lib.LineData('key  # comment', 'devices', True)
        assert False
    except ValueError:
        pass


def test_enabled():
    values = (('yes', True), ('1', True), ('no', False), ('0', False))
    for value, is_enabled in values:
        data = lib.LineData('key ' + value, 'defaults', False)
        assert data.type == data.TYPE_OPTION
        assert data.option == 'key'
        assert data.value == value
        assert data.is_enabled() == is_enabled

    data = lib.LineData('key neither_yes_nor_no', 'defaults', False)
    assert data.type == data.TYPE_OPTION
    assert data.option == 'key'
    assert data.value == 'neither_yes_nor_no'
    assert data.is_enabled() is None



from leapp.libraries.actor import iscmodel
from leapp.libraries.common import isccfg
from leapp.models import BindFacts


def model_paths(issues_model):
    paths = list()
    for m in issues_model:
        paths.append(m.path)
    return paths


def get_facts(cfg):
    facts = iscmodel.get_facts(cfg)
    assert isinstance(facts, BindFacts)
    return facts


def test_simple():
    mockcfg = isccfg.MockConfig("""
options {
    listen-on port 53 { 127.0.0.1; };
    listen-on-v6 port 53 { ::1; };
    directory       "/var/named";
    allow-query     { localhost; };
    recursion yes;

    dnssec-validation yes;
};

zone "." IN {
    type hint;
    file "named.ca";
};
""", '/etc/named.conf')
    facts = get_facts(mockcfg)
    assert facts.dnssec_lookaside is None


def test_dnssec_lookaside():
    mockcfg = isccfg.MockConfig("""
options {
    listen-on port 53 { 127.0.0.1; };
    listen-on-v6 port 53 { ::1; };
    directory       "/var/named";
    allow-query     { localhost; };
    recursion yes;

    dnssec-validation yes;
        dnssec-lookaside auto;
};

zone "." IN {
    type hint;
    file "named.ca";
};
""", '/etc/named.conf')
    facts = get_facts(mockcfg)
    assert '/etc/named.conf' in model_paths(facts.dnssec_lookaside)


def test_listen_on_v6():
    present = isccfg.MockConfig("""
options {
    listen-on { any; };
    listen-on-v6 { any; };
};
""", '/etc/named.conf')
    missing = isccfg.MockConfig("""
options {
    listen-on { any; };
    #listen-on-v6 { any; };
};
""", '/etc/named.conf')

    facts = get_facts(present)
    assert not facts.listen_on_v6_missing

    facts = get_facts(missing)
    assert facts.listen_on_v6_missing



from leapp import reporting
from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import architecture
from leapp.models import FirmwareFacts, GrubInfo
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag

GRUB_SUMMARY = ('On legacy (BIOS) systems, GRUB2 core (located in the gap between the MBR and the '
                'first partition) cannot be updated during the rpm transaction and Leapp has to initiate '
                'the update running "grub2-install" after the transaction. No action is needed before the '
                'upgrade. After the upgrade, it is recommended to check the GRUB configuration.')


class CheckGrubCore(Actor):
    """
    Check whether we are on legacy (BIOS) system and instruct Leapp to upgrade GRUB core
    """

    name = 'check_grub_core'
    consumes = (FirmwareFacts, GrubInfo)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if architecture.matches_architecture(architecture.ARCH_S390X):
            # s390x archs use ZIPL instead of GRUB
            return

        ff = next(self.consume(FirmwareFacts), None)
        if ff and ff.firmware == 'bios':
            grub_info = next(self.consume(GrubInfo), None)
            if not grub_info:
                raise StopActorExecutionError('Actor did not receive any GrubInfo message.')
            if grub_info.orig_devices:
                create_report([
                    reporting.Title(
                        'GRUB2 core will be automatically updated during the upgrade'
                    ),
                    reporting.Summary(GRUB_SUMMARY),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.BOOT]),
                ])
            else:
                create_report([
                    reporting.Title('Leapp could not identify where GRUB2 core is located'),
                    reporting.Summary(
                        'We assumed GRUB2 core is located on the same device(s) as /boot, '
                        'however Leapp could not detect GRUB2 on the device(s). '
                        'GRUB2 core needs to be updated maually on legacy (BIOS) systems. '
                    ),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.BOOT]),
                    reporting.Remediation(
                        hint='Please run "grub2-install <GRUB_DEVICE> command manually after the upgrade'),
                ])



import os

import pytest

from leapp.models import SelinuxRelabelDecision
from leapp.snactor.fixture import current_actor_context

# TODO These tests modifies the system


@pytest.mark.skipif(
    os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
    reason='Test disabled by default because it would modify the system',
)
def test_schedule_no_relabel(current_actor_context):
    current_actor_context.feed(SelinuxRelabelDecision(set_relabel=False))
    current_actor_context.run()
    assert not os.path.isfile('/.autorelabel')


@pytest.mark.skipif(
    os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
    reason='Test disabled by default because it would modify the system',
)
def test_schedule_relabel(current_actor_context):
    current_actor_context.feed(SelinuxRelabelDecision(set_relabel=True))
    current_actor_context.run()
    assert os.path.isfile('/.autorelabel')

    # lets cleanup so we possibly not affect further testing
    os.unlink('/.autorelabel')



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class SysctlVariable(Model):
    topic = SystemFactsTopic

    name = fields.String()
    value = fields.String()


class SysctlVariablesFacts(Model):
    topic = SystemFactsTopic

    sysctl_variables = fields.List(fields.Model(SysctlVariable))



from leapp.actors import Actor
from leapp.libraries.actor.ziplcheckbootentries import inhibit_if_invalid_zipl_configuration
from leapp.models import SourceBootLoaderConfiguration
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class ZiplCheckBootEntries(Actor):
    """
    Inhibits the upgrade if a problematic Zipl configuration is detected on the system.

    The configuration is considered problematic if it will cause troubles during its conversion to BLS.
    Such troubles can be caused by either containing multiple rescue entries, or containing rescue entries
    sharing the same kernel image version.
    """

    name = 'zipl_check_boot_entries'
    consumes = (SourceBootLoaderConfiguration,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        boot_loader_configuration = next(self.consume(SourceBootLoaderConfiguration))
        inhibit_if_invalid_zipl_configuration(boot_loader_configuration)



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class PkgManagerInfo(Model):
    """
    Package manager (yum/dnf) related info

    We expect to have only one single message of this kind produced
    """
    topic = SystemInfoTopic

    etc_releasever = fields.Nullable(fields.String())
    """
    Contain the first line of /etc/{yum,dnf}/vars/releasever file or None if the file does not exist.

    In case the value is empty string, it means the file exists but it is empty. In such a case the
    original configuration is obviously broken.
    """

    configured_proxies = fields.List(fields.String(), default=[])
    """
    A sorted list of proxies present in yum and dnf configuration files.
    """

    enabled_plugins = fields.List(fields.String(), default=[])



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class PowerTop(Actor):
    """
    Check if PowerTOP is installed. If yes, write information about non-compatible changes.
    """

    name = 'powertop'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'powertop'):
            create_report([
                reporting.Title('PowerTOP compatibility options removed in the next major version'),
                reporting.Summary(
                    'The -d (dump) option which has been kept for RHEL backward compatibility has been '
                    'dropped.\n'
                    'The -h option which has been used for RHEL backward compatibility is no longer '
                    'alias for --html, but it\'s now an alias for --help to follow the upstream.\n'
                    'The -u option which has been used for RHEL backward compatibility as an alias for '
                    '--help has been dropped.\n'
                ),
                reporting.Severity(reporting.Severity.LOW),
                reporting.Groups([reporting.Groups.TOOLS, reporting.Groups.MONITORING]),
                reporting.Remediation(hint='Please remove the dropped options from your scripts.'),
                reporting.RelatedResource('package', 'powertop')
            ])



from leapp import reporting
from leapp.libraries.common.config import architecture, version
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, TrackedFilesInfoSource

DEFAULT_OPENSSL_CONF = '/etc/pki/tls/openssl.cnf'
URL_8_CRYPTOPOLICIES = 'https://red.ht/rhel-8-system-wide-crypto-policies'
URL_9_CRYPTOPOLICIES = 'https://red.ht/rhel-9-system-wide-crypto-policies'


def check_ibmca():
    if not architecture.matches_architecture(architecture.ARCH_S390X):
        # not needed check really, but keeping it to make it clear
        return
    if not has_package(DistributionSignedRPM, 'openssl-ibmca'):
        return
    # In RHEL 9 has been introduced new technology: openssl providers. The engine
    # is deprecated, so keep proper teminology to not confuse users.
    dst_tech = 'engine' if version.get_target_major_version() == '8' else 'providers'
    summary = (
        'The presence of openssl-ibmca package suggests that the system may be configured'
        ' to use the IBMCA OpenSSL engine.'
        ' Due to major changes in OpenSSL and libica between RHEL {source} and RHEL {target} it is not'
        ' possible to migrate OpenSSL configuration files automatically. Therefore,'
        ' it is necessary to enable IBMCA {tech} in the OpenSSL config file manually'
        ' after the system upgrade.'
        .format(
            source=version.get_source_major_version(),
            target=version.get_target_major_version(),
            tech=dst_tech
        )
    )

    hint = (
        'Configure the IBMCA {tech} manually after the upgrade.'
        ' Please, be aware that it is not recommended to configure the system default'
        ' {fpath}. Instead, it is recommended to configure a copy of'
        ' that file and use this copy only for particular applications that are supposed'
        ' to utilize the IBMCA {tech}. The location of the OpenSSL configuration file'
        ' can be specified using the OPENSSL_CONF environment variable.'
        .format(tech=dst_tech, fpath=DEFAULT_OPENSSL_CONF)
    )

    reporting.create_report([
        reporting.Title('Detected possible use of IBMCA in OpenSSL'),
        reporting.Summary(summary),
        reporting.Remediation(hint=hint),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([
            reporting.Groups.POST,
            reporting.Groups.ENCRYPTION
        ]),
    ])


def _is_openssl_modified():
    tracked_files = next(api.consume(TrackedFilesInfoSource), None)
    if not tracked_files:
        # unexpected at all, skipping testing, but keeping the log just in case
        api.current_logger.warning('The TrackedFilesInfoSource message is missing! Skipping check of openssl config.')
        return False
    for finfo in tracked_files.files:
        if finfo.path == DEFAULT_OPENSSL_CONF:
            return finfo.is_modified
    return False


def check_default_openssl():
    if not _is_openssl_modified():
        return

    crypto_url = URL_8_CRYPTOPOLICIES if version.get_target_major_version == '8' else URL_9_CRYPTOPOLICIES

    # TODO(pstodulk): Needs in future some rewording, as OpenSSL engines are
    # deprecated since "RHEL 8" and people should use OpenSSL providers instead.
    # (IIRC, they are required to use OpenSSL providers since RHEL 9.) The
    # current wording could be inaccurate.
    summary = (
        'The OpenSSL configuration file ({fpath}) has been'
        ' modified on the system. RHEL 8 (and newer) systems provide a crypto-policies'
        ' mechanism ensuring usage of system-wide secure cryptography algorithms.'
        ' Also the target system uses newer version of OpenSSL that is not fully'
        ' compatible with the current one.'
        ' To ensure the upgraded system uses crypto-policies as expected,'
        ' the new version of the openssl configuration file must be installed'
        ' during the upgrade. This will be done automatically.'
        ' The original configuration file will be saved'
        ' as "{fpath}.leappsave".'
        '\n\nNote this can affect the ability to connect to the system after'
        ' the upgrade if it depends on the current OpenSSL configuration.'
        ' Such a problem may be caused by using a particular OpenSSL engine, as'
        ' OpenSSL engines built for the'
        ' RHEL {source} system are not compatible with RHEL {target}.'
        .format(
            fpath=DEFAULT_OPENSSL_CONF,
            source=version.get_source_major_version(),
            target=version.get_target_major_version()
        )
    )
    if version.get_target_major_version() == '9':
        # NOTE(pstodulk): that a try to make things with engine/providers a
        # little bit better (see my TODO note above)
        summary += (
            '\n\nNote the legacy ENGINE API is deprecated since RHEL 8 and'
            ' it is required to use the new OpenSSL providers API instead on'
            ' RHEL 9 systems.'
        )
    hint = (
        'Check that your ability to login to the system does not depend on'
        ' the OpenSSL configuration. After the upgrade, review the system configuration'
        ' and configure the system as needed.'
        ' Please, be aware that it is not recommended to configure the system default'
        ' {fpath}. Instead, it is recommended to copy the file and use this copy'
        ' to configure particular applications.'
        ' The default OpenSSL configuration file should be modified only'
        ' when it is really necessary.'
    )
    reporting.create_report([
        reporting.Title('The /etc/pki/tls/openssl.cnf file is modified and will be replaced during the upgrade.'),
        reporting.Summary(summary),
        reporting.Remediation(hint=hint),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.POST, reporting.Groups.SECURITY]),
        reporting.RelatedResource('file', DEFAULT_OPENSSL_CONF),
        reporting.ExternalLink(
            title='Using system-wide cryptographic policies.',
            url=crypto_url
        )
    ])


def process():
    check_ibmca()
    check_default_openssl()



import textwrap

from mock import patch

from leapp.libraries.actor.authselectscanner import Authconfig, AuthselectScannerLibrary, DConf, read_file
from leapp.libraries.common.pam import PAM


def get_config(config):
    return textwrap.dedent(config).strip()


def test_read_file__non_existent():
    content = read_file('/this/does/not/exist')
    assert content == ''


def test_read_file__ok():
    content = read_file(__file__)
    assert content != ''
    assert 'test_read_file__ok' in content


def test_Authconfig_get_bool__non_existent():
    obj = Authconfig('')
    assert not obj.get_bool('non-existent-option')


def test_Authconfig_get_bool__true():
    obj = Authconfig(get_config('''
    test_a=True
    test_b=true
    test_c=Yes
    test_d=yes
    '''))

    assert obj.get_bool('test_a')
    assert obj.get_bool('test_b')
    assert obj.get_bool('test_c')
    assert obj.get_bool('test_d')


def test_Authconfig_get_bool__false():
    obj = Authconfig(get_config('''
    test_a=False
    test_b=false
    test_c=No
    test_d=no
    '''))

    assert not obj.get_bool('test_a')
    assert not obj.get_bool('test_b')
    assert not obj.get_bool('test_c')
    assert not obj.get_bool('test_d')


def test_Authconfig_get_string__non_existent():
    obj = Authconfig('')
    assert obj.get_string('non-existent-option') is None


def test_Authconfig_get_string__ok():
    obj = Authconfig(get_config('''
    test_a="str"
    test_b=str
    '''))

    assert obj.get_string('test_a') == 'str'
    assert obj.get_string('test_b') == 'str'


def test_DConf_get_bool__non_existent():
    obj = DConf('')
    assert not obj.get_bool('section', 'non-existent-option')


def test_DConf_get_bool__true():
    obj = DConf(get_config('''
    [section]
    test_a=True
    test_b=true
    test_c=Yes
    test_d=yes
    '''))

    assert obj.get_bool('section', 'test_a')
    assert obj.get_bool('section', 'test_b')
    assert obj.get_bool('section', 'test_c')
    assert obj.get_bool('section', 'test_d')


def test_DConf_get_bool__false():
    obj = DConf(get_config('''
    [section]
    test_a=False
    test_b=false
    test_c=No
    test_d=no
    '''))

    assert not obj.get_bool('section', 'test_a')
    assert not obj.get_bool('section', 'test_b')
    assert not obj.get_bool('section', 'test_c')
    assert not obj.get_bool('section', 'test_d')


def test_DConf_get_string__non_existent():
    obj = DConf('')
    assert obj.get_string('section', 'non-existent-option') is None


def test_DConf_get_string__ok():
    obj = DConf(get_config('''
    [section]
    test_a="str"
    test_b=str
    '''))

    assert obj.get_string('section', 'test_a') == 'str'
    assert obj.get_string('section', 'test_b') == 'str'


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__None(mock_service):
    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(''), '')
    mock_service.return_value = False
    assert obj.step_detect_profile() is None


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__sssd(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = False
    assert obj.step_detect_profile() == 'sssd'


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__winbind(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_winbind.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = False
    assert obj.step_detect_profile() == 'winbind'


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__nis(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = True
    assert obj.step_detect_profile() == 'nis'


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__sssd_winbind(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_winbind.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = False
    assert obj.step_detect_profile() is None


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__sssd_nis(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = True
    assert obj.step_detect_profile() is None


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__winbind_nis(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_winbind.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = True
    assert obj.step_detect_profile() is None


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
def test_AuthselectScannerLibrary_step_detect_profile__sssd_winbind_nis(mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_winbind.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    mock_service.return_value = True
    assert obj.step_detect_profile() is None


def test_AuthselectScannerLibrary_step_detect_features__faillock():
    pam = get_config('''
    auth required pam_faillock.so preauth silent deny=4 unlock_time=1200
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    assert obj.step_detect_features() == ['with-faillock']


def test_AuthselectScannerLibrary_step_detect_features__fingerprint():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth sufficient pam_fprintd.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    assert obj.step_detect_features() == ['with-fingerprint']


def test_AuthselectScannerLibrary_step_detect_features__access():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    account required pam_access.so
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    assert obj.step_detect_features() == ['with-pamaccess']


def test_AuthselectScannerLibrary_step_detect_features__mkhomedir():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    session optional pam_mkhomedir.so umask=0077
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    assert obj.step_detect_features() == ['with-mkhomedir']


def test_AuthselectScannerLibrary_step_detect_features__mkhomedir_oddjob():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    session optional pam_oddjob_mkhomedir.so umask=0077
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    assert obj.step_detect_features() == ['with-mkhomedir']


def test_AuthselectScannerLibrary_step_detect_features__all():
    pam = get_config('''
    auth required pam_faillock.so preauth silent deny=4 unlock_time=1200
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth sufficient pam_fprintd.so
    auth required pam_deny.so
    account required pam_access.so
    session optional pam_oddjob_mkhomedir.so umask=0077
    ''')

    obj = AuthselectScannerLibrary([], Authconfig(''), DConf(''), PAM(pam), '')
    features = obj.step_detect_features()
    assert len(features) == 4
    assert 'with-faillock' in features
    assert 'with-fingerprint' in features
    assert 'with-pamaccess' in features
    assert 'with-mkhomedir' in features


def test_AuthselectScannerLibrary_step_detect_sssd_features__sudo():
    nsswitch = get_config('''
    passwd:     files sss systemd
    group:      files sss systemd
    sudoers:    files sss
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), nsswitch
    )
    features = obj.step_detect_sssd_features('sssd')
    assert features == ['with-sudo']


def test_AuthselectScannerLibrary_step_detect_sssd_features__smartcard():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    ac = get_config('''
    USESMARTCARD=yes
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(ac), DConf(''), PAM(pam), ''
    )
    features = obj.step_detect_sssd_features('sssd')
    assert features == ['with-smartcard']


def test_AuthselectScannerLibrary_step_detect_sssd_features__smartcard_required():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    ac = get_config('''
    FORCESMARTCARD=yes
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(ac), DConf(''), PAM(pam), ''
    )
    features = obj.step_detect_sssd_features('sssd')
    assert features == ['with-smartcard-required']


def test_AuthselectScannerLibrary_step_detect_sssd_features__smartcard_lock():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    dconf = get_config('''
    [org/gnome/settings-daemon/peripherals/smartcard]
    removal-action='lock-screen'
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(dconf), PAM(pam), ''
    )
    features = obj.step_detect_sssd_features('sssd')
    assert features == ['with-smartcard-lock-on-removal']


def test_AuthselectScannerLibrary_step_detect_sssd_features__pkcs11():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_pkcs11.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    ac = get_config('''
    USESMARTCARD=yes
    FORCESMARTCARD=yes
    ''')

    dconf = get_config('''
    [org/gnome/settings-daemon/peripherals/smartcard]
    removal-action='lock-screen'
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(ac), DConf(dconf), PAM(pam), ''
    )
    features = obj.step_detect_sssd_features('sssd')
    assert not features


def test_AuthselectScannerLibrary_step_detect_sssd_features__wrong_profile():
    nsswitch = get_config('''
    passwd:     files sss systemd
    group:      files sss systemd
    sudoers:    files sss
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), nsswitch
    )
    features = obj.step_detect_sssd_features('winbind')
    assert not features


def test_AuthselectScannerLibrary_step_detect_winbind_features__krb5():
    ac = get_config('''
    WINBINDKRB5=yes
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(ac), DConf(''), PAM(''), ''
    )
    features = obj.step_detect_winbind_features('winbind')
    assert features == ['with-krb5']


def test_AuthselectScannerLibrary_step_detect_winbind_features__wrong_profile():
    ac = get_config('''
    WINBINDKRB5=yes
    ''')

    obj = AuthselectScannerLibrary(
        [], Authconfig(ac), DConf(''), PAM(''), ''
    )
    features = obj.step_detect_winbind_features('sssd')
    assert not features


@patch('os.readlink')
@patch('os.path.islink')
@patch('os.path.isfile')
@patch('os.path.getmtime')
def test_AuthselectScannerLibrary_step_detect_if_confirmation_is_required__nosysconfig(
    mock_getmtime, mock_isfile, mock_islink, mock_readlink
):
    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), ''
    )
    mock_isfile.return_value = False
    assert obj.step_detect_if_confirmation_is_required()


@patch('os.readlink')
@patch('os.path.islink')
@patch('os.path.isfile')
@patch('os.path.getmtime')
def test_AuthselectScannerLibrary_step_detect_if_confirmation_is_required__nolink(
    mock_getmtime, mock_isfile, mock_islink, mock_readlink
):
    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), ''
    )
    mock_isfile.return_value = True
    mock_islink.return_value = False
    assert obj.step_detect_if_confirmation_is_required()


@patch('os.readlink')
@patch('os.path.islink')
@patch('os.path.isfile')
@patch('os.path.getmtime')
def test_AuthselectScannerLibrary_step_detect_if_confirmation_is_required__badlink(
    mock_getmtime, mock_isfile, mock_islink, mock_readlink
):
    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), ''
    )
    mock_isfile.return_value = True
    mock_islink.return_value = True
    mock_readlink.return_value = ''
    assert obj.step_detect_if_confirmation_is_required()


@patch('os.readlink')
@patch('os.path.islink')
@patch('os.path.isfile')
@patch('os.path.getmtime')
def test_AuthselectScannerLibrary_step_detect_if_confirmation_is_required__badmtime(
    mock_getmtime, mock_isfile, mock_islink, mock_readlink
):
    def my_getmtime(path):
        # Make sysconfig file older then other files.
        if path == '/etc/sysconfig/authconfig':
            return 100

        return 200

    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), ''
    )
    mock_isfile.return_value = True
    mock_islink.return_value = True
    mock_readlink.side_effect = '{}-ac'.format
    mock_getmtime.side_effect = my_getmtime
    assert obj.step_detect_if_confirmation_is_required()


@patch('os.readlink')
@patch('os.path.islink')
@patch('os.path.isfile')
@patch('os.path.getmtime')
def test_AuthselectScannerLibrary_step_detect_if_confirmation_is_required__pass(
    mock_getmtime, mock_isfile, mock_islink, mock_readlink
):
    def my_getmtime(path):
        # Make sysconfig file younger then other files.
        if path == '/etc/sysconfig/authconfig':
            return 200

        return 100

    obj = AuthselectScannerLibrary(
        [], Authconfig(''), DConf(''), PAM(''), ''
    )
    mock_isfile.return_value = True
    mock_islink.return_value = True
    mock_readlink.side_effect = '{}-ac'.format
    mock_getmtime.side_effect = my_getmtime
    assert not obj.step_detect_if_confirmation_is_required()


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
@patch('leapp.libraries.actor.authselectscanner.AuthselectScannerLibrary.step_detect_if_confirmation_is_required')
def test_AuthselectScannerLibrary_process__simple(mock_confirm, mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary(
        ['pam_unix', 'pam_sss', 'pam_deny'], Authconfig(''), DConf(''), PAM(pam), ''
    )
    mock_confirm.return_value = True
    mock_service.return_value = False
    authselect = obj.process()
    assert authselect.profile == 'sssd'
    assert not authselect.features
    assert authselect.confirm


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
@patch('leapp.libraries.actor.authselectscanner.AuthselectScannerLibrary.step_detect_if_confirmation_is_required')
def test_AuthselectScannerLibrary_process__features(mock_confirm, mock_service):
    pam = get_config('''
    auth required pam_faillock.so preauth silent deny=4 unlock_time=1200
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    nsswitch = get_config('''
    passwd:     files sss systemd
    group:      files sss systemd
    sudoers:    files sss
    ''')

    obj = AuthselectScannerLibrary(
        ['pam_unix', 'pam_sss', 'pam_deny', 'pam_faillock'],
        Authconfig(''),
        DConf(''),
        PAM(pam),
        nsswitch
    )
    mock_confirm.return_value = True
    mock_service.return_value = False
    authselect = obj.process()
    assert authselect.profile == 'sssd'
    assert len(authselect.features) == 2
    assert 'with-faillock' in authselect.features
    assert 'with-sudo' in authselect.features
    assert authselect.confirm


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
@patch('leapp.libraries.actor.authselectscanner.AuthselectScannerLibrary.step_detect_if_confirmation_is_required')
def test_AuthselectScannerLibrary_process__unknown_module(mock_confirm, mock_service):
    pam = get_config('''
    auth required pam_faillock.so preauth silent deny=4 unlock_time=1200
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary(
        ['pam_unix', 'pam_sss', 'pam_deny'],
        Authconfig(''),
        DConf(''),
        PAM(pam),
        ''
    )
    mock_confirm.return_value = True
    mock_service.return_value = False
    authselect = obj.process()
    assert authselect.profile is None
    assert not authselect.features
    assert authselect.confirm


@patch('leapp.libraries.actor.authselectscanner.is_service_enabled')
@patch('leapp.libraries.actor.authselectscanner.AuthselectScannerLibrary.step_detect_if_confirmation_is_required')
def test_AuthselectScannerLibrary_process__autoconfirm(mock_confirm, mock_service):
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = AuthselectScannerLibrary(
        ['pam_unix', 'pam_sss', 'pam_deny'], Authconfig(''), DConf(''), PAM(pam), ''
    )
    mock_confirm.return_value = False
    mock_service.return_value = False
    authselect = obj.process()
    assert authselect.profile == 'sssd'
    assert not authselect.features
    assert not authselect.confirm



import os
import platform

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import CalledProcessError, run
from leapp.models import EnvVar, IPUConfig, OSRelease, Version

ENV_IGNORE = ('LEAPP_CURRENT_PHASE', 'LEAPP_CURRENT_ACTOR', 'LEAPP_VERBOSE',
              'LEAPP_DEBUG')

ENV_MAPPING = {'LEAPP_DEVEL_DM_DISABLE_UDEV': 'DM_DISABLE_UDEV'}


def get_env_vars():
    """
    Gather LEAPP_DEVEL environment variables and respective mappings to provide them as messages to be
    available after reboot.
    """
    env_vars = []
    leapp_vars = {k: v for (k, v) in os.environ.items() if k.startswith('LEAPP_') and k not in ENV_IGNORE}
    for k, v in leapp_vars.items():
        if k in ENV_MAPPING:
            env_vars.append(EnvVar(name=ENV_MAPPING.get(k), value=v))
            continue
        env_vars.append(EnvVar(name=k, value=v))

    return env_vars


def get_booted_kernel():
    """
    Get version and release of the currently used kernel in one string.
    """
    try:
        return run(['/usr/bin/uname', '-r'])['stdout'].strip()
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='Unable to obtain release of the booted kernel.',
            details={'details': str(e), 'stderr': e.stderr}
        )


def get_os_release(path):
    """
    Retrieve data about System OS release from provided file.

    :return: `OSRelease` model if the file can be parsed
    :raises: `IOError`
    """
    try:
        with open(path) as f:
            data = dict(l.strip().split('=', 1) for l in f.readlines() if '=' in l)
            return OSRelease(
                release_id=data.get('ID', '').strip('"'),
                name=data.get('NAME', '').strip('"'),
                pretty_name=data.get('PRETTY_NAME', '').strip('"'),
                version=data.get('VERSION', '').strip('"'),
                version_id=data.get('VERSION_ID', '').strip('"'),
                variant=data.get('VARIANT', '').strip('"') or None,
                variant_id=data.get('VARIANT_ID', '').strip('"') or None
            )
    except IOError as e:
        raise StopActorExecutionError(message='Cannot collect the system OS facts.',
                                      details={'details': str(e)})


def produce_ipu_config(actor):
    flavour = os.environ.get('LEAPP_UPGRADE_PATH_FLAVOUR')
    target_version = os.environ.get('LEAPP_UPGRADE_PATH_TARGET_RELEASE')
    os_release = get_os_release('/etc/os-release')

    actor.produce(IPUConfig(
        leapp_env_vars=get_env_vars(),
        os_release=os_release,
        architecture=platform.machine(),
        version=Version(
            source=os_release.version_id,
            target=target_version
        ),
        kernel=get_booted_kernel(),
        flavour=flavour
    ))



from leapp.actors import Actor
from leapp.libraries.actor import sanebackendsmigrate
from leapp.models import DistributionSignedRPM
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class SanebackendsMigrate(Actor):
    """
    Actor for migrating sane-backends configuration files.

    Adds USB quirks for support specific USB scanners if they
    are not added during package manager transaction.
    """

    name = 'sanebackends_migrate'
    consumes = (DistributionSignedRPM,)
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        sanebackendsmigrate.update_sane()



from leapp import reporting
from leapp.libraries.common.config import get_env
from leapp.reporting import create_report


def checkcifs(storage_info):
    # if network in initramfs is enabled CIFS inhibitor is redundant
    if get_env('LEAPP_DEVEL_INITRAM_NETWORK', None):
        return
    for storage in storage_info:
        if any(entry.fs_vfstype == "cifs" for entry in storage.fstab):
            create_report([
                reporting.Title("Use of CIFS detected. Upgrade can't proceed"),
                reporting.Summary("CIFS is currently not supported by the inplace upgrade."),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups([
                        reporting.Groups.FILESYSTEM,
                        reporting.Groups.NETWORK
                ]),
                reporting.Remediation(hint='Comment out CIFS entries to proceed with the upgrade.'),
                reporting.Groups([reporting.Groups.INHIBITOR]),
                reporting.RelatedResource('file', '/etc/fstab')
            ])



#
# Helper functions
#

from os.path import isfile

from leapp.libraries.actor import sctpdlm
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import ActiveKernelModulesFacts


def anyfile(files):
    """
    Determines if any of the given paths exist and are a file.

    :type files: tuple of str
    :return: True if any of the given paths exists and it is a file.
    :rtype: bool
    """
    for f in files:
        try:
            if isfile(f):
                return True
        except OSError:
            continue
    return False


def is_module_loaded(module):
    """
    Determines if the given kernel module has been reported in the ActiveKernelModuleFacts as loaded.

    :return: True if the module has been found in the ActiveKernelModuleFacts.
    :rtype: bool
    """
    for fact in api.consume(ActiveKernelModulesFacts):
        for active_module in fact.kernel_modules:
            if active_module.filename == module:
                return True
    return False


def is_sctp_used():
    """
    Logic function that decides whether SCTP is being used on this machine.

    :return: True if SCTP usage was detected.
    :rtype: bool
    """

    # If anything is using SCTP, be it for listening on new connections or
    # connecting somewhere else, the module will be loaded. Thus, no need to
    # also probe on sockets.
    if is_module_loaded('sctp'):
        return True

    # Basic files from lksctp-tools. This check is enough and checking RPM
    # database is an overkill here and this allows for checking for
    # manually installed ones, which is possible.
    lksctp_files = ['/usr/lib64/libsctp.so.1',
                    '/usr/lib/libsctp.so.1',
                    '/usr/bin/sctp_test']
    if anyfile(lksctp_files):
        api.current_logger().debug('At least one of lksctp files is present.')
        return True

    if sctpdlm.is_dlm_using_sctp():
        return True

    return False


def was_sctp_used():
    """
    Determines whether SCTP has been used in the path month, by checking the journalctl.

    :return: True if SCTP usage has been found.
    :rtype: bool
    """
    try:
        run(['check_syslog_for_sctp.sh'])
    except CalledProcessError:
        api.current_logger().debug('Nothing regarding SCTP was found on journal.')
        return False
    api.current_logger().debug('Found logs regarding SCTP on journal.')
    return True


def is_sctp_wanted():
    """
    Decision making function that decides based on the current or past usage of SCTP, the SCTP module is wanted
    on the new system.

    :return: True if SCTP seems to be in use or has been recently used.
    :rtype: bool
    """
    if is_sctp_used():
        api.current_logger().info('SCTP is being used.')
        return True

    if was_sctp_used():
        api.current_logger().info('SCTP was used.')
        return True

    api.current_logger().info('SCTP is not being used and neither wanted.')
    return False



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class Module(Model):
    """
    A single DNF module identified by its name and stream.
    """
    topic = SystemFactsTopic
    name = fields.String()
    stream = fields.String()


class EnabledModules(Model):
    """
    DNF modules enabled on the source system.
    """
    topic = SystemFactsTopic

    modules = fields.List(fields.Model(Module), default=[])



from leapp import reporting

LIST_SEPARATOR_FMT = '\n    - '


def inhibit_if_deprecated_directives_used(ssh_config_msg):
    """ Inhibits the upgrade if any deprecated directives were found in the sshd configuration. """

    if ssh_config_msg.deprecated_directives:
        # Prepare the output of the deprecated directives for the user
        deprecated_directives_report_text = ''
        for deprecated_directive in ssh_config_msg.deprecated_directives:
            deprecated_directives_report_text += '{0}{1}'.format(LIST_SEPARATOR_FMT, deprecated_directive)

        sshd_config_path = '/etc/ssh/sshd_config'
        reporting.create_report([
            reporting.Title('A deprecated directive in the sshd configuration'),
            reporting.Summary(
                'The following deprecated directives were found in the sshd configuration file {0}:{1}'
                .format(sshd_config_path, deprecated_directives_report_text)
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SERVICES]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.RelatedResource('file', sshd_config_path),
            reporting.Remediation(
                hint='Remove the deprecated directives from the sshd configuration.',
                commands=[["sed", "-i", "/^\\s*ShowPatchLevel.*$/Id", sshd_config_path]]
            )
        ])



import os

from leapp.libraries.actor import multipathconfread
from leapp.models import MultipathConfig8to9

TEST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')


def build_config(pathname, config_dir, enable_foreign_exists, invalid_regexes_exist, allow_usb_exists):
    return MultipathConfig8to9(
        pathname=pathname,
        config_dir=config_dir,
        enable_foreign_exists=enable_foreign_exists,
        invalid_regexes_exist=invalid_regexes_exist,
        allow_usb_exists=allow_usb_exists,
    )


def assert_config(config, expected):
    assert config.pathname == expected.pathname
    assert config.config_dir == expected.config_dir
    assert config.enable_foreign_exists == expected.enable_foreign_exists
    assert config.invalid_regexes_exist == expected.invalid_regexes_exist
    assert config.allow_usb_exists == expected.allow_usb_exists


default_rhel8_conf = build_config(
    os.path.join(TEST_DIR, 'default_rhel8.conf'), None, True, False, False)

all_the_things_conf = build_config(
    os.path.join(TEST_DIR, 'all_the_things.conf'), None, False, True, False)

converted_the_things_conf = build_config(
    os.path.join(TEST_DIR, 'converted_the_things.conf'), None, True, False, True)

complicated_conf = build_config(
    os.path.join(TEST_DIR, 'complicated.conf'), "/etc/multipath/conf.d", True, True, False)

no_foreign_conf = build_config(
    os.path.join(TEST_DIR, 'no_foreign.conf'), None, False, True, True)

allow_usb_conf = build_config(
    os.path.join(TEST_DIR, 'allow_usb.conf'), None, False, False, True)

empty_conf = build_config(
    os.path.join(TEST_DIR, 'empty.conf'), None, False, False, False)

missing_dir_conf = build_config(
    os.path.join(TEST_DIR, 'missing_dir.conf'), os.path.join(TEST_DIR, 'missing'), False, True, False)

empty_dir_conf = build_config(
    os.path.join(TEST_DIR, 'empty_dir.conf'), os.path.join(TEST_DIR, 'conf3.d'), False, False, False)

not_set_dir_conf = build_config(
    os.path.join(TEST_DIR, 'not_set_dir.conf'), os.path.join(TEST_DIR, "conf1.d"), False, True, False)

empty1_conf = build_config(
    os.path.join(TEST_DIR, 'conf1.d/empty.conf'), None, False, False, False)

nothing_important_conf = build_config(
    os.path.join(TEST_DIR, 'conf1.d/nothing_important.conf'),
    os.path.join(TEST_DIR, 'this_gets_ignored'), False, False, False)

set_in_dir_conf = build_config(
    os.path.join(TEST_DIR, 'set_in_dir.conf'), os.path.join(TEST_DIR, "conf2.d"), False, False, False)

all_true_conf = build_config(
    os.path.join(TEST_DIR, 'conf2.d/all_true.conf'), None, True, True, True)

no_defaults_conf = build_config(
    os.path.join(TEST_DIR, 'no_defaults.conf'), None, False, True, False)

two_defaults_conf = build_config(
    os.path.join(TEST_DIR, 'two_defaults.conf'), None, True, False, False)


def mock_parse_config(path):
    """Convert config_dir into full pathname"""
    conf = multipathconfread._parse_config_orig(path)
    if not conf:
        return None
    if conf.config_dir:
        conf.config_dir = os.path.join(TEST_DIR, conf.config_dir)
    return conf


def test_parse_config():
    test_map = {'default_rhel8.conf': default_rhel8_conf,
                'all_the_things.conf': all_the_things_conf,
                'converted_the_things.conf': converted_the_things_conf,
                'complicated.conf': complicated_conf,
                'no_foreign.conf': no_foreign_conf,
                'allow_usb.conf': allow_usb_conf,
                'no_defaults.conf': no_defaults_conf,
                'two_defaults.conf': two_defaults_conf,
                'empty.conf': empty_conf}
    for config_name, expected_data in test_map.items():
        config = multipathconfread._parse_config(os.path.join(TEST_DIR, config_name))
        assert config
        assert_config(config, expected_data)


def test_get_facts_missing_dir(monkeypatch):
    monkeypatch.setattr(multipathconfread, '_parse_config_orig', multipathconfread._parse_config, raising=False)
    monkeypatch.setattr(multipathconfread, '_parse_config', mock_parse_config)

    facts = multipathconfread.get_multipath_conf_facts(os.path.join(TEST_DIR, 'missing_dir.conf'))
    assert facts
    assert len(facts.configs) == 1
    assert_config(facts.configs[0], missing_dir_conf)


def test_get_facts_empty_dir(monkeypatch):
    monkeypatch.setattr(multipathconfread, '_parse_config_orig', multipathconfread._parse_config, raising=False)
    monkeypatch.setattr(multipathconfread, '_parse_config', mock_parse_config)

    facts = multipathconfread.get_multipath_conf_facts(os.path.join(TEST_DIR, 'empty_dir.conf'))
    assert facts
    assert len(facts.configs) == 1
    assert_config(facts.configs[0], empty_dir_conf)


def test_get_facts_not_set_dir(monkeypatch):
    monkeypatch.setattr(multipathconfread, '_parse_config_orig', multipathconfread._parse_config, raising=False)
    monkeypatch.setattr(multipathconfread, '_parse_config', mock_parse_config)

    expected_configs = (not_set_dir_conf, empty1_conf, nothing_important_conf)
    facts = multipathconfread.get_multipath_conf_facts(os.path.join(TEST_DIR, 'not_set_dir.conf'))
    assert facts
    assert len(facts.configs) == 3
    for i in range(len(facts.configs)):
        assert_config(facts.configs[i], expected_configs[i])


def test_get_facts_set_in_dir(monkeypatch):
    monkeypatch.setattr(multipathconfread, '_parse_config_orig', multipathconfread._parse_config, raising=False)
    monkeypatch.setattr(multipathconfread, '_parse_config', mock_parse_config)

    expected_configs = (set_in_dir_conf, all_true_conf)
    facts = multipathconfread.get_multipath_conf_facts(os.path.join(TEST_DIR, 'set_in_dir.conf'))
    assert facts
    assert len(facts.configs) == 2
    for i in range(len(facts.configs)):
        assert_config(facts.configs[i], expected_configs[i])



from leapp.actors import Actor
from leapp.libraries.actor import rocecheck
from leapp.models import KernelCmdline, Report, RoceDetected
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class RoceCheck(Actor):
    """
    Check whether RoCE is used on the system and well configured for the upgrade.

    This is valid only for IBM Z systems (s390x). If a used RoCE is detected,
        * system must be RHEL 8.7+ (suggesting 8.8+ due to 8.7 EOL)
        * and system must be booted with: net.naming-scheme=rhel-8.7
    otherwise the network is broken due to changed NICs.
    """

    name = 'roce_check'
    consumes = (KernelCmdline, RoceDetected)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        rocecheck.process()



from leapp.actors import Actor
from leapp.libraries.actor import selinuxprepare
from leapp.models import SELinuxCustom, SELinuxModules
from leapp.tags import IPUWorkflowTag, PreparationPhaseTag


class SELinuxPrepare(Actor):
    """
    Remove selinux policy customizations before updating selinux-policy* packages

    Policy customizations in the original system could cause policy package
    upgrade to fail and therefore need to be removed.
    Customizations introduced by semanage are removed first, followed by custom
    policy modules gathered by SELinuxContentScanner.
    """

    name = 'selinuxprepare'
    consumes = (SELinuxCustom, SELinuxModules)
    produces = ()
    tags = (PreparationPhaseTag, IPUWorkflowTag)

    def process(self):
        selinuxprepare.remove_semanage_customizations()
        selinuxprepare.remove_custom_modules()



from leapp.libraries.stdlib import api
from leapp.models import ActiveKernelModulesFacts, DetectedDeviceOrDriver, DeviceDriverDeprecationData


def process():
    loaded_drivers = {
        module.filename
        for message in api.consume(ActiveKernelModulesFacts)
        for module in message.kernel_modules
    }
    driver_data = {
        entry.driver_name: entry
        for message in api.consume(DeviceDriverDeprecationData)
        for entry in message.entries
        if not entry.device_id
    }
    api.produce(*[
        DetectedDeviceOrDriver(**driver_data[driver].dump())
        for driver in loaded_drivers
        if driver in driver_data
    ])



from leapp.actors import Actor
from leapp.libraries.common import persistentnetnames
from leapp.models import PersistentNetNamesFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class PersistentNetNames(Actor):
    """
    Get network interface information for physical ethernet interfaces of the original system.

    Gather information like PCI topology, MAC address and interface name on the
    original system. Those data are provided through the PersistentNetNamesFacts
    model.

    See the persistentnetnamesinitramfs actor that is very same but processed
    during early phases in initrams to gather the same data but using already
    new kernel of the target system to be able to reflect changes affected
    by the new kernel (see the PersistentNetNamesConfig actor).
    """

    name = 'persistentnetnames'
    consumes = ()
    produces = (PersistentNetNamesFacts, )
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        self.produce(PersistentNetNamesFacts(interfaces=list(persistentnetnames.interfaces())))



import pytest

from leapp.libraries.actor import scanclienablerepo
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import CustomTargetRepository


class LoggerMocked(object):
    def __init__(self):
        self.infomsg = None
        self.debugmsg = None

    def info(self, msg):
        self.infomsg = msg

    def debug(self, msg):
        self.debugmsg = msg

    def __call__(self):
        return self


def test_no_enabledrepos(monkeypatch):
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'current_logger', LoggerMocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    scanclienablerepo.process()
    assert not api.current_logger.infomsg
    assert not api.produce.called


@pytest.mark.parametrize('envars,result', [
    ({'LEAPP_ENABLE_REPOS': 'repo1'}, [CustomTargetRepository(repoid='repo1')]),
    ({'LEAPP_ENABLE_REPOS': 'repo1,repo2'}, [CustomTargetRepository(repoid='repo1'),
                                             CustomTargetRepository(repoid='repo2')]),
])
def test_enabledrepos(monkeypatch, envars, result):
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'current_logger', LoggerMocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(envars=envars))
    scanclienablerepo.process()
    assert api.current_logger.infomsg
    assert api.produce.called == len(result)
    for i in result:
        assert i in api.produce.model_instances



from leapp.libraries.actor import quaggadaemons
from leapp.models import QuaggaToFrrFacts

# daemons for mocked _check_service function
TEST_DAEMONS = ['bgpd', 'ospfd', 'zebra']


def mock_check_service(name, state):
    if name in TEST_DAEMONS:
        return True

    return False


def test_process_daemons():
    quaggadaemons._check_service = mock_check_service

    facts = quaggadaemons.process_daemons()
    assert isinstance(facts, QuaggaToFrrFacts)
    assert facts.active_daemons == TEST_DAEMONS
    assert facts.enabled_daemons == TEST_DAEMONS



from leapp.actors import Actor
from leapp.messaging.commands import SkipPhasesUntilCommand
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckSkipPhase(Actor):
    """
    Skip all the subsequent phases until the report phase.

    The phases that follow after the Checks phase work with the target
    RHEL major version user space - stuff around preparing and checking
    the rpm transaction. We do not want to process those phases in case of
    inhibition - e.g. for a specific HW unsupported by the target system we
    cannot do anything - we can just see some unclear errors in such case.
    So we want to instead skip to the Reports phase to provide clear report
    to user without confusing errors.

    The actor is processed after all actors in the phase (that provides Report
    messages) are processed.
    """

    name = 'check_skip_phase'
    consumes = (Report,)
    produces = ()
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        results = list(self.consume(Report))
        inhibitors = [msg for msg in results if 'inhibitor' in msg.report.get('flags', [])]
        if inhibitors:
            self.log.info("An upgrade inhibitor detected. Skipping to the Report phase.")
            # until_phase='targettransactioncheck'  === phase after this phase will be processed
            # === the Reports phase
            self._messaging.command(SkipPhasesUntilCommand(until_phase='targettransactioncheck'))



import itertools
import os
import re
import shutil

from leapp import reporting
from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp.libraries.actor import constants
from leapp.libraries.common import dnfplugin, mounting, overlaygen, repofileutils, rhsm, utils
from leapp.libraries.common.config import get_env, get_product_type
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.common.gpg import get_path_to_gpg_certs, is_nogpgcheck_set
from leapp.libraries.stdlib import api, CalledProcessError, config, run
from leapp.models import RequiredTargetUserspacePackages  # deprecated
from leapp.models import TMPTargetRepositoriesFacts  # deprecated all the time
from leapp.models import (
    CustomTargetRepositoryFile,
    PkgManagerInfo,
    RepositoriesFacts,
    RHSMInfo,
    RHUIInfo,
    StorageInfo,
    TargetOSInstallationImage,
    TargetRepositories,
    TargetUserSpaceInfo,
    TargetUserSpacePreupgradeTasks,
    UsedTargetRepositories,
    UsedTargetRepository,
    XFSPresence
)
from leapp.utils.deprecation import suppress_deprecation

# TODO: "refactor" (modify) the library significantly
# The current shape is really bad and ineffective (duplicit parsing
# of repofiles). The library is doing 3 (5) things:
# # (0.) consume process input data
# # 1. prepare the first container, to be able to obtain repositories for the
# #    target system (this is extra neededwhen rhsm is used, but not reason to
# #    do such thing only when rhsm is used. Be persistent here
# # 2. gather target repositories that should AND can be used
# #    - basically here is the main thing that is PITA; I started
# #      the refactoring but realized that it needs much more changes because
# #      of RHSM...
# # 3. create the target userspace bootstrap
# # (4.) produce messages with the data
#
# Because of the lack of time, I am extending the current bad situation,
# but after the release, the related code should be really refactored.
# It would be probably ideal, if this and other actors in the current and the
# next phase are modified properly and we could create inhibitors in the check
# phase and keep everything on the report. But currently it seems it doesn't
# worth to invest so much energy into it. So let's just make this really
# readable (includes split of the functionality into several libraries)
# and do not mess.
# Issue: #486

PROD_CERTS_FOLDER = 'prod-certs'
PERSISTENT_PACKAGE_CACHE_DIR = '/var/lib/leapp/persistent_package_cache'
DEDICATED_LEAPP_PART_URL = 'https://access.redhat.com/solutions/7011704'


def _check_deprecated_rhsm_skip():
    # we do not plan to cover this case by tests as it is purely
    # devel/testing stuff, that becomes deprecated now
    # just log the warning now (better than nothing?); deprecation process will
    # be specified in close future
    if get_env('LEAPP_DEVEL_SKIP_RHSM', '0') == '1':
        api.current_logger().warning(
            'The LEAPP_DEVEL_SKIP_RHSM has been deprecated. Use'
            ' LEAPP_NO_RHSM instead or use the --no-rhsm option for'
            ' leapp. as well custom repofile has not been defined.'
            ' Please read documentation about new "skip rhsm" solution.'
        )


class BrokenSymlinkError(Exception):
    """Raised when we encounter a broken symlink where we weren't expecting it."""


class _InputData(object):
    def __init__(self):
        self._consume_data()

    @suppress_deprecation(RequiredTargetUserspacePackages)
    def _consume_data(self):
        """
        Wrapper function to consume majority input data.

        It doesn't consume TargetRepositories, which are consumed in the
        own function.
        """
        self.packages = {'dnf', 'dnf-command(config-manager)'}
        self.files = []
        _cftuples = set()

        def _update_files(copy_files):
            # add just uniq CopyFile objects to omit duplicate copying of files
            for cfile in copy_files:
                cftuple = (cfile.src, cfile.dst)
                if cftuple not in _cftuples:
                    _cftuples.add(cftuple)
                    self.files.append(cfile)

        for task in api.consume(TargetUserSpacePreupgradeTasks):
            self.packages.update(task.install_rpms)
            _update_files(task.copy_files)

        for message in api.consume(RequiredTargetUserspacePackages):
            self.packages.update(message.packages)

        # Get the RHSM information (available repos, attached SKUs, etc.) of the source system
        self.rhsm_info = next(api.consume(RHSMInfo), None)
        self.rhui_info = next(api.consume(RHUIInfo), None)
        if not self.rhsm_info and not rhsm.skip_rhsm():
            api.current_logger().warning('Could not receive RHSM information - Is this system registered?')
            raise StopActorExecution()
        if rhsm.skip_rhsm() and self.rhsm_info:
            # this should not happen. if so, raise an error as something in
            # other actors is wrong really
            raise StopActorExecutionError("RHSM is not handled but the RHSMInfo message has been produced.")

        self.custom_repofiles = list(api.consume(CustomTargetRepositoryFile))
        self.xfs_info = next(api.consume(XFSPresence), XFSPresence())
        self.storage_info = next(api.consume(StorageInfo), None)
        if not self.storage_info:
            raise StopActorExecutionError('No storage info available cannot proceed.')


def _restore_persistent_package_cache(userspace_dir):
    if get_env('LEAPP_DEVEL_USE_PERSISTENT_PACKAGE_CACHE', None) == '1':
        if not os.path.exists(PERSISTENT_PACKAGE_CACHE_DIR):
            return
        dst_cache = os.path.join(userspace_dir, 'var', 'cache', 'dnf')
        if os.path.exists(dst_cache):
            run(['rm', '-rf', dst_cache])
        shutil.move(PERSISTENT_PACKAGE_CACHE_DIR, dst_cache)
    # We always want to remove the persistent cache here to unclutter the system
    run(['rm', '-rf', PERSISTENT_PACKAGE_CACHE_DIR])


def _backup_to_persistent_package_cache(userspace_dir):
    if get_env('LEAPP_DEVEL_USE_PERSISTENT_PACKAGE_CACHE', None) == '1':
        # Clean up any dead bodies, just in case
        run(['rm', '-rf', PERSISTENT_PACKAGE_CACHE_DIR])
        src_cache = os.path.join(userspace_dir, 'var', 'cache', 'dnf')
        if os.path.exists(src_cache):
            shutil.move(src_cache, PERSISTENT_PACKAGE_CACHE_DIR)


def _import_gpg_keys(context, install_root_dir, target_major_version):
    certs_path = get_path_to_gpg_certs()
    # Import the RHEL X+1 GPG key to be able to verify the installation of initial packages
    try:
        # Import also any other keys provided by the customer in the same directory
        for certname in os.listdir(certs_path):
            cmd = ['rpm', '--root', install_root_dir, '--import', os.path.join(certs_path, certname)]
            context.call(cmd, callback_raw=utils.logging_handler)
    except CalledProcessError as exc:
        raise StopActorExecutionError(
            message=(
                'Unable to import GPG certificates to install RHEL {} userspace packages.'
                .format(target_major_version)
            ),
            details={'details': str(exc), 'stderr': exc.stderr}
        )


def _handle_transaction_err_msg_size_old(err):
    # NOTE(pstodulk): This is going to be removed in future!

    article_section = 'Generic case'
    xfs_info = next(api.consume(XFSPresence), XFSPresence())
    if xfs_info.present and xfs_info.without_ftype:
        article_section = 'XFS ftype=0 case'

    message = ('There is not enough space on the file system hosting /var/lib/leapp directory '
               'to extract the packages.')
    details = {'hint': "Please follow the instructions in the '{}' section of the article at: "
                       "link: https://access.redhat.com/solutions/5057391".format(article_section)}

    raise StopActorExecutionError(message=message, details=details)


def _handle_transaction_err_msg_size(err):
    if get_env('LEAPP_OVL_LEGACY', '0') == '1':
        _handle_transaction_err_msg_size_old(err)
        return  # not needed actually as the above function raises error, but for visibility
    NO_SPACE_STR = 'more space needed on the'

    # Disk Requirements:
    #   At least <size> more space needed on the <path> filesystem.
    #
    missing_space = [line.strip() for line in err.stderr.split('\n') if NO_SPACE_STR in line]
    size_str = re.match(r'At least (.*) more space needed', missing_space[0]).group(1)
    message = 'There is not enough space on the file system hosting /var/lib/leapp.'
    hint = (
        'Increase the free space on the filesystem hosting'
        ' /var/lib/leapp by {} at minimum. It is suggested to provide'
        ' reasonably more space to be able to perform all planned actions'
        ' (e.g. when 200MB is missing, add 1700MB or more).\n\n'
        'It is also a good practice to create dedicated partition'
        ' for /var/lib/leapp when more space is needed, which can be'
        ' dropped after the system upgrade is fully completed'
        ' For more info, see: {}'
        .format(size_str, DEDICATED_LEAPP_PART_URL)
    )
    # we do not want to confuse customers by the orig msg speaking about
    # missing space on '/'. Skip the Disk Requirements section.
    # The information is part of the hint.
    details = {'hint': hint}

    raise StopActorExecutionError(message=message, details=details)


def prepare_target_userspace(context, userspace_dir, enabled_repos, packages):
    """
    Implement the creation of the target userspace.
    """
    _backup_to_persistent_package_cache(userspace_dir)

    run(['rm', '-rf', userspace_dir])
    _create_target_userspace_directories(userspace_dir)

    target_major_version = get_target_major_version()
    install_root_dir = '/el{}target'.format(target_major_version)
    with mounting.BindMount(source=userspace_dir, target=os.path.join(context.base_dir, install_root_dir.lstrip('/'))):
        _restore_persistent_package_cache(userspace_dir)
        if not is_nogpgcheck_set():
            _import_gpg_keys(context, install_root_dir, target_major_version)

        repos_opt = [['--enablerepo', repo] for repo in enabled_repos]
        repos_opt = list(itertools.chain(*repos_opt))
        cmd = ['dnf', 'install', '-y']
        if is_nogpgcheck_set():
            cmd.append('--nogpgcheck')
        cmd += [
            '--setopt=module_platform_id=platform:el{}'.format(target_major_version),
            '--setopt=keepcache=1',
            '--releasever', api.current_actor().configuration.version.target,
            '--installroot', install_root_dir,
            '--disablerepo', '*'
            ] + repos_opt + packages
        if config.is_verbose():
            cmd.append('-v')
        if rhsm.skip_rhsm():
            cmd += ['--disableplugin', 'subscription-manager']
        try:
            context.call(cmd, callback_raw=utils.logging_handler)
        except CalledProcessError as exc:
            message = 'Unable to install RHEL {} userspace packages.'.format(target_major_version)
            details = {'details': str(exc), 'stderr': exc.stderr}

            if 'more space needed on the' in exc.stderr:
                # The stderr contains this error summary:
                # Disk Requirements:
                #   At least <size> more space needed on the <path> filesystem.
                _handle_transaction_err_msg_size(exc)

            # If a proxy was set in dnf config, it should be the reason why dnf
            # failed since leapp does not support updates behind proxy yet.
            for manager_info in api.consume(PkgManagerInfo):
                if manager_info.configured_proxies:
                    details['details'] = (
                        "DNF failed to install userspace packages, likely due to the proxy "
                        "configuration detected in the YUM/DNF configuration file. "
                        "Make sure the proxy is properly configured in /etc/dnf/dnf.conf. "
                        "It's also possible the proxy settings in the DNF configuration file are "
                        "incompatible with the target system. A compatible configuration can be "
                        "placed in /etc/leapp/files/dnf.conf which, if present, will be used during "
                        "the upgrade instead of /etc/dnf/dnf.conf. "
                        "In such case the configuration will also be applied to the target system."
                    )

            # Similarly if a proxy was set specifically for one of the repositories.
            for repo_facts in api.consume(RepositoriesFacts):
                for repo_file in repo_facts.repositories:
                    if any(repo_data.proxy and repo_data.enabled for repo_data in repo_file.data):
                        details['details'] = (
                            "DNF failed to install userspace packages, likely due to the proxy "
                            "configuration detected in a repository configuration file."
                        )

            raise StopActorExecutionError(message=message, details=details)


def _query_rpm_for_pkg_files(context, pkgs):
    files_owned_by_rpm = set()
    rpm_query_result = context.call(['rpm', '-ql'] + pkgs, split=True)
    files_owned_by_rpm.update(rpm_query_result['stdout'])
    return files_owned_by_rpm


def _get_files_owned_by_rpms(context, dirpath, pkgs=None, recursive=False):
    """
    Return the list of file names inside dirpath owned by RPMs.

    This is important e.g. in case of RHUI which installs specific repo files
    in the yum.repos.d directory.

    In case the pkgs param is None or empty, do not filter any specific rpms.
    Otherwise return filenames that are owned by any pkg in the given list.

    If the recursive param is set to True, all files owned by a package in the
    directory tree starting at dirpath are returned. Otherwise, only the
    files within dirpath are checked.
    """

    files_owned_by_rpms = []

    file_list = []
    searchdir = context.full_path(dirpath)
    if recursive:
        for root, _, files in os.walk(searchdir):
            for filename in files:
                relpath = os.path.relpath(os.path.join(root, filename), searchdir)
                file_list.append(relpath)
    else:
        file_list = os.listdir(searchdir)

    for fname in file_list:
        try:
            result = context.call(['rpm', '-qf', os.path.join(dirpath, fname)])
        except CalledProcessError:
            api.current_logger().debug('SKIP the {} file: not owned by any rpm'.format(fname))
            continue
        if pkgs and not [pkg for pkg in pkgs if pkg in result['stdout']]:
            api.current_logger().debug('SKIP the {} file: not owned by any searched rpm:'.format(fname))
            continue
        api.current_logger().debug('Found the file owned by an rpm: {}.'.format(fname))
        files_owned_by_rpms.append(fname)

    return files_owned_by_rpms


def _mkdir_with_copied_mode(path, mode_from):
    """
    Create directories with a file to copy the mode from.

    :param path: The directory path to create.
    :param mode_from: A file or directory whose mode we will copy to the
        newly created directory.
    :raises subprocess.CalledProcessError: mkdir or chmod fails. For instance,
        the directory already exists, the file to get permissions from does
        not exist, a parent directory does not exist.
    """
    # Create with maximally restrictive permissions
    run(['mkdir', '-m', '0', '-p', path])
    run(['chmod', '--reference={}'.format(mode_from), path])


def _choose_copy_or_link(symlink, srcdir):
    """
    Determine whether to copy file contents or create a symlink depending on where the pointee resides.

    :param symlink: The source symlink to follow.  This must be an absolute path.
    :param srcdir: The root directory that every piece of content must be present in.
    :returns: A tuple of action and sourcefile.  Action is one of 'copy' or 'link' and means that
        the caller should either copy the sourcefile to the target location or create a symlink from
        the sourcefile to the target location.  sourcefile is the path to the file that should be
        the source of the operation.  It is either a real file outside of the srcdir hierarchy or
        a file (real, directory, symlink or otherwise) inside of the srcdir hierarchy.
    :raises ValueError: if the arguments are not correct
    :raises BrokenSymlinkError: if the symlink is invalid

    Determine whether the file pointed to by the symlink chain is within srcdir.  If it is within,
    then create a synlink that points from symlink to it.

    If it is not within, then walk the symlink chain until we find something that is within srcdir
    and return that. This means we will omit any symlinks that are outside of srcdir from
    the symlink chain.

    If we reach a real file and it is outside of srcdir, then copy the file instead.
    """
    if not symlink.startswith('/'):
        raise ValueError('File{} must be an absolute path!'.format(symlink))

    # os.path.exists follows symlinks
    if not os.path.exists(symlink):
        raise BrokenSymlinkError('File {} is a broken symlink!'.format(symlink))

    # If srcdir is a symlink, then we need a name for it that we can compare
    # with other paths.
    canonical_srcdir = os.path.realpath(srcdir)

    pointee_as_abspath = symlink
    seen = set([pointee_as_abspath])

    # The goal of this while loop is to find the next link in a possible
    # symlink chain that either points to a symlink inside of srcdir or to
    # a file or directory that we can copy.
    while os.path.islink(pointee_as_abspath):
        # Advance pointee to the target of the previous link
        pointee = os.readlink(pointee_as_abspath)

        # Note: os.path.join()'s behaviour if the pointee is an absolute path
        # essentially ignores the first argument (which is what we want).
        pointee_as_abspath = os.path.normpath(os.path.join(os.path.dirname(pointee_as_abspath), pointee))

        # Make sure we aren't in a circular set of references.
        # On Linux, this should not happen as the os.path.exists() call
        # before the loop should catch it but we don't want to enter an
        # infinite loop if that code changes later.
        if pointee_as_abspath in seen:
            if symlink == pointee_as_abspath:
                error_msg = ('File {} is a broken symlink that references'
                             ' itself!'.format(pointee_as_abspath))
            else:
                error_msg = ('File {} references {} which is a broken symlink'
                             ' that references itself!'.format(symlink, pointee_as_abspath))

            raise BrokenSymlinkError(error_msg)

        seen.add(pointee_as_abspath)

        # To make comparisons, we need to resolve all symlinks in the directory
        # structure leading up to pointee.  However, we can't include pointee
        # itself otherwise it will resolve to the file that it points to in the
        # end (which would be wrong if pointee_filename is a symlink).
        canonical_pointee_dir, pointee_filename = os.path.split(pointee_as_abspath)
        canonical_pointee_dir = os.path.realpath(canonical_pointee_dir)

        if canonical_pointee_dir.startswith(canonical_srcdir):
            # Absolute path inside of the correct dir so we need to link to it
            # But we need to determine what the link path should be before
            # returning.

            # Construct a relative path that points from the symlinks directory
            # to the pointee.
            link_to = os.readlink(symlink)
            canonical_symlink_dir = os.path.realpath(os.path.dirname(symlink))
            relative_path = os.path.relpath(canonical_pointee_dir, canonical_symlink_dir)

            if link_to.startswith('/'):
                # The original symlink was an absolute path so we will set this
                # one to absolute too
                # Note: Because absolute paths are constructed inside of
                # srcdir, the relative path that we need to join here has to be
                # relative to srcdir, not the directory that the symlink is
                # being created in.
                relative_to_srcdir = os.path.relpath(canonical_pointee_dir, canonical_srcdir)
                corrected_path = os.path.normpath(os.path.join(srcdir, relative_to_srcdir, pointee_filename))

            else:
                # If the original link is a relative link, then we want the new
                # link to be relative as well
                corrected_path = os.path.normpath(os.path.join(relative_path, pointee_filename))

            return ("link", corrected_path)

        # pointee is a symlink that points outside of the srcdir so continue to
        # the next symlink in the chain.

    # The file is not a link so copy it
    return ('copy', pointee_as_abspath)


def _copy_symlinks(symlinks_to_process, srcdir):
    """
    Copy file contents or create a symlink depending on where the pointee resides.

    :param symlinks_to_process: List of 2-tuples of (src_path, target_path).  Each src_path
        should be an absolute path to the symlink.  target_path is the path to where we
        need to create either a link or a copy.
    :param srcdir: The root directory that every piece of content must be present in.
    :raises ValueError: if the arguments are not correct
    """
    for source_linkpath, target_linkpath in symlinks_to_process:
        try:
            action, source_path = _choose_copy_or_link(source_linkpath, srcdir)
        except BrokenSymlinkError as e:
            # Skip and report broken symlinks
            api.current_logger().warning('{} Will not copy the file!'.format(str(e)))
            continue

        if action == "copy":
            # Note: source_path could be a directory, so '-a' or '-r' must be
            # given to cp.
            run(['cp', '-a', source_path, target_linkpath])
        elif action == 'link':
            run(["ln", "-s", source_path, target_linkpath])
        else:
            # This will not happen unless _copy_or_link() has a bug.
            raise RuntimeError("Programming error: _copy_or_link() returned an unknown action:{}".format(action))


def _copy_decouple(srcdir, dstdir):
    """
    Copy files inside of `srcdir` to `dstdir` while decoupling symlinks.

    What we mean by decoupling the `srcdir` is that any symlinks pointing
    outside the directory will be copied as regular files. This means that the
    directory will become independent from its surroundings with respect to
    symlinks. Any symlink (or symlink chains) within the directory will be
    preserved.

    .. warning::
        `dstdir` must already exist.
    """
    for root, directories, files in os.walk(srcdir):
        # relative path from srcdir because srcdir is replaced with dstdir for
        # the copy.
        relpath = os.path.relpath(root, srcdir)

        # Create all directories with proper permissions for security
        # reasons (Putting private data into directories that haven't had their
        # permissions set appropriately may leak the private information.)
        symlinks_to_process = []
        for directory in directories:
            source_dirpath = os.path.join(root, directory)
            target_dirpath = os.path.join(dstdir, relpath, directory)

            # Defer symlinks until later because we may end up having to copy
            # the file contents and the directory may not exist yet.
            if os.path.islink(source_dirpath):
                symlinks_to_process.append((source_dirpath, target_dirpath))
                continue

            _mkdir_with_copied_mode(target_dirpath, source_dirpath)

        # Link or create all directories that were pointed to by symlinks and
        # then reset symlinks_to_process for use by files.
        _copy_symlinks(symlinks_to_process, srcdir)
        symlinks_to_process = []

        for filename in files:
            source_filepath = os.path.join(root, filename)
            target_filepath = os.path.join(dstdir, relpath, filename)

            # Defer symlinks until later because we may end up having to copy
            # the file contents and the directory may not exist yet.
            if os.path.islink(source_filepath):
                symlinks_to_process.append((source_filepath, target_filepath))
                continue

            # Not a symlink so we can copy it now too
            run(['cp', '-a', source_filepath, target_filepath])

        _copy_symlinks(symlinks_to_process, srcdir)


def _copy_certificates(context, target_userspace):
    """
    Copy certificates from source system into the container, but preserve
    original ones

    Some certificates are already installed in the container and those are
    default certificates for the target OS, so we preserve these.

    We respect the symlink hierarchy of the source system within the /etc/pki
    folder. Dangling symlinks will be ignored.

    """

    target_pki = os.path.join(target_userspace, 'etc', 'pki')
    backup_pki = os.path.join(target_userspace, 'etc', 'pki.backup')

    with mounting.NspawnActions(base_dir=target_userspace) as target_context:
        files_owned_by_rpms = _get_files_owned_by_rpms(target_context, '/etc/pki', recursive=True)
        api.current_logger().debug('Files owned by rpms: {}'.format(' '.join(files_owned_by_rpms)))

    # Backup container /etc/pki
    run(['mv', target_pki, backup_pki])

    # _copy_decouple() requires we create the target_pki directory here because we don't know
    # the mode inside of _copy_decouple().
    _mkdir_with_copied_mode(target_pki, backup_pki)

    # Copy source /etc/pki to the container
    _copy_decouple('/etc/pki', target_pki)

    # Assertion: after running _copy_decouple(), no broken symlinks exist in /etc/pki in the container
    # So any broken symlinks created will be by the installed packages.

    # Recover installed packages as they always get precedence
    for filepath in files_owned_by_rpms:
        src_path = os.path.join(backup_pki, filepath)
        dst_path = os.path.join(target_pki, filepath)

        # Resolve and skip any broken symlinks
        is_broken_symlink = False
        pointee = None
        if os.path.islink(src_path):
            pointee = os.path.join(target_userspace, os.readlink(src_path)[1:])

            seen = set()
            while os.path.islink(pointee):
                # The symlink points to a path relative to the target userspace so
                # we need to readjust it
                pointee = os.path.join(target_userspace, os.readlink(src_path)[1:])
                if not os.path.exists(pointee) or pointee in seen:
                    is_broken_symlink = True

                    # The path original path of the broken symlink in the container
                    report_path = os.path.join(target_pki, os.path.relpath(src_path, backup_pki))
                    api.current_logger().warning(
                            'File {} is a broken symlink! Will not copy!'.format(report_path))
                    break

                seen.add(pointee)

        if is_broken_symlink:
            continue

        # Cleanup conflicting files
        run(['rm', '-rf', dst_path])

        # Ensure destination exists
        parent_dir = os.path.dirname(dst_path)
        run(['mkdir', '-p', parent_dir])

        # Copy the new file
        run(['cp', '-R', '--preserve=all', src_path, dst_path])

    run(['rm', '-rf', backup_pki])


def _prep_repository_access(context, target_userspace):
    """
    Prepare repository access by copying all relevant certificates and configuration files to the userspace
    """
    target_etc = os.path.join(target_userspace, 'etc')
    target_yum_repos_d = os.path.join(target_etc, 'yum.repos.d')
    backup_yum_repos_d = os.path.join(target_etc, 'yum.repos.d.backup')

    _copy_certificates(context, target_userspace)
    # NOTE(dkubek): context.call(['update-ca-trust']) seems to not be working.
    #               I am not really sure why. The changes to files are not
    #               being written to disk.
    run(["chroot", target_userspace, "/bin/bash", "-c", "su - -c update-ca-trust"])

    if not rhsm.skip_rhsm():
        run(['rm', '-rf', os.path.join(target_etc, 'rhsm')])
        context.copytree_from('/etc/rhsm', os.path.join(target_etc, 'rhsm'))

    # NOTE: We cannot just remove the target yum.repos.d dir and replace it with yum.repos.d from the scratch
    # #     that we've used to obtain the new DNF stack and install it into the target userspace. Although
    # #     RHUI clients are being installed in both scratch and target containers, users can request their package
    # #     to be installed into target userspace that might add some repos to yum.repos.d that are not in scratch.

    # Detect files that are owned by some RPM - these cannot be deleted
    with mounting.NspawnActions(base_dir=target_userspace) as target_context:
        files_owned_by_rpms = _get_files_owned_by_rpms(target_context, '/etc/yum.repos.d')

    # Backup the target yum.repos.d so we can always copy the files installed by some RPM back into yum.repos.d
    # when we modify it
    run(['mv', target_yum_repos_d, backup_yum_repos_d])

    # Copy the yum.repos.d from scratch - preserve any custom repositories. No need to clean-up old RHUI clients,
    # we swap them for the new RHUI client in scratch (so the old one is not installed).
    context.copytree_from('/etc/yum.repos.d', target_yum_repos_d)

    # Copy back files owned by some RPM
    for fname in files_owned_by_rpms:
        api.current_logger().debug('Copy the backed up repo file: {}'.format(fname))
        run(['mv', os.path.join(backup_yum_repos_d, fname), os.path.join(target_yum_repos_d, fname)])

    # Cleanup - remove the backed up dir
    run(['rm', '-rf', backup_yum_repos_d])


def _get_product_certificate_path():
    """
    Retrieve the required / used product certificate for RHSM.
    """
    architecture = api.current_actor().configuration.architecture
    target_version = api.current_actor().configuration.version.target
    target_product_type = get_product_type('target')
    certs_dir = api.get_common_folder_path(PROD_CERTS_FOLDER)

    # We do not need any special certificates to reach repos from non-ga channels, only beta requires special cert.
    if target_product_type != 'beta':
        target_product_type = 'ga'

    prod_certs = {
        'x86_64': {
            'ga': '479.pem',
            'beta': '486.pem',
        },
        'aarch64': {
            'ga': '419.pem',
            'beta': '363.pem',
        },
        'ppc64le': {
            'ga': '279.pem',
            'beta': '362.pem',
        },
        's390x': {
            'ga': '72.pem',
            'beta': '433.pem',
        }
    }

    try:
        cert = prod_certs[architecture][target_product_type]
    except KeyError as e:
        raise StopActorExecutionError(message='Failed to determine what certificate to use for {}.'.format(e))

    cert_path = os.path.join(certs_dir, target_version, cert)
    if not os.path.isfile(cert_path):
        additional_summary = ''
        if target_product_type != 'ga':
            additional_summary = (
                ' This can happen when upgrading a beta system and the chosen target version does not have'
                ' beta certificates attached (for example, because the GA has been released already).'

            )

        reporting.create_report([
            reporting.Title('Cannot find the product certificate file for the chosen target system.'),
            reporting.Summary(
                'Expected certificate: {cert} with path {path} but it could not be found.{additional}'.format(
                    cert=cert, path=cert_path, additional=additional_summary)
            ),
            reporting.Groups([reporting.Groups.REPOSITORY]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Remediation(hint=(
                'Set the corresponding target os version in the LEAPP_DEVEL_TARGET_RELEASE environment variable for'
                'which the {cert} certificate is provided'.format(cert=cert)
            )),
        ])
        raise StopActorExecution()

    return cert_path


def _create_target_userspace_directories(target_userspace):
    api.current_logger().debug('Creating target userspace directories.')
    try:
        utils.makedirs(target_userspace)
        api.current_logger().debug('Done creating target userspace directories.')
    except OSError:
        api.current_logger().error(
            'Failed to create temporary target userspace directories %s', target_userspace, exc_info=True)
        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {directory} is empty and modifiable.'.format(directory=target_userspace)
            }
        )


def _inhibit_on_duplicate_repos(repofiles):
    """
    Inhibit the upgrade if any repoid is defined multiple times.

    When that happens, it not only shows misconfigured system, but then
    we can't get details of all the available repos as well.
    """
    # TODO: this is is duplicate of rhsm._inhibit_on_duplicate_repos
    # Issue: #486
    duplicates = repofileutils.get_duplicate_repositories(repofiles).keys()

    if not duplicates:
        return
    list_separator_fmt = '\n    - '
    api.current_logger().warning(
        'The following repoids are defined multiple times:{0}{1}'
        .format(list_separator_fmt, list_separator_fmt.join(duplicates))
    )

    reporting.create_report([
        reporting.Title('A YUM/DNF repository defined multiple times'),
        reporting.Summary(
            'The following repositories are defined multiple times inside the'
            ' "upgrade" container:{0}{1}'
            .format(list_separator_fmt, list_separator_fmt.join(duplicates))
        ),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.REPOSITORY]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.Remediation(hint=(
            'Remove the duplicate repository definitions or change repoids of'
            ' conflicting repositories on the system to prevent the'
            ' conflict.'
            )
        )
    ])


def _get_all_available_repoids(context):
    repofiles = repofileutils.get_parsed_repofiles(context)
    # TODO: this is not good solution, but keep it as it is now
    # Issue: #486
    if rhsm.skip_rhsm():
        # only if rhsm is skipped, the duplicate repos are not detected
        # automatically and we need to do it extra
        _inhibit_on_duplicate_repos(repofiles)
    repoids = []
    for rfile in repofiles:
        if rfile.data:
            repoids += [repo.repoid for repo in rfile.data]
    return set(repoids)


def _get_rhsm_available_repoids(context):
    target_major_version = get_target_major_version()
    # FIXME: check that required repo IDs (baseos, appstream)
    # + or check that all required RHEL repo IDs are available.
    if rhsm.skip_rhsm():
        return set()
    # Get the RHSM repos available in the target RHEL container
    # TODO: very similar thing should happens for all other repofiles in container
    #
    repoids = rhsm.get_available_repo_ids(context)
    # NOTE(ivasilev) For the moment at least AppStream and BaseOS repos are required. While we are still
    # contemplating on what can be a generic solution to checking this, let's introduce a minimal check for
    # at-least-one-appstream and at-least-one-baseos among present repoids
    if not repoids or all("baseos" not in ri for ri in repoids) or all("appstream" not in ri for ri in repoids):
        reporting.create_report([
            reporting.Title('Cannot find required basic RHEL target repositories.'),
            reporting.Summary(
                'This can happen when a repository ID was entered incorrectly either while using the --enablerepo'
                ' option of leapp or in a third party actor that produces a CustomTargetRepositoryMessage.'
            ),
            reporting.Groups([reporting.Groups.REPOSITORY]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Remediation(hint=(
                'It is required to have RHEL repositories on the system'
                ' provided by the subscription-manager unless the --no-rhsm'
                ' option is specified. You might be missing a valid SKU for'
                ' the target system or have a failed network connection.'
                ' Check whether your system is attached to a valid SKU that is'
                ' providing RHEL {} repositories.'
                ' If you are using Red Hat Satellite, read the upgrade documentation'
                ' to set up Satellite and the system properly.'

            ).format(target_major_version)),
            reporting.ExternalLink(
                # https://red.ht/preparing-for-upgrade-to-rhel8
                # https://red.ht/preparing-for-upgrade-to-rhel9
                # https://red.ht/preparing-for-upgrade-to-rhel10
                url='https://red.ht/preparing-for-upgrade-to-rhel{}'.format(target_major_version),
                title='Preparing for the upgrade')
            ])
        raise StopActorExecution()
    return set(repoids)


def _get_rhui_available_repoids(context, cloud_repo):
    repofiles = repofileutils.get_parsed_repofiles(context)

    # TODO: same refactoring as Issue #486?
    _inhibit_on_duplicate_repos(repofiles)
    repoids = []
    for rfile in repofiles:
        if rfile.file == cloud_repo and rfile.data:
            repoids = [repo.repoid for repo in rfile.data]
            repoids.sort()
            break
    return set(repoids)


def get_copy_location_from_copy_in_task(context, copy_task):
    basename = os.path.basename(copy_task.src)
    dest_in_container = context.full_path(copy_task.dst)
    if os.path.isdir(dest_in_container):
        return os.path.join(copy_task.dst, basename)
    return copy_task.dst


def _get_rh_available_repoids(context, indata):
    """
    RH repositories are provided either by RHSM or are stored in the expected repo file provided by
    RHUI special packages (every cloud provider has itw own rpm).
    """

    rh_repoids = _get_rhsm_available_repoids(context)

    # If we are upgrading a RHUI system, check what repositories are provided by the (already installed) target clients
    if indata and indata.rhui_info:
        files_provided_by_clients = _query_rpm_for_pkg_files(context, indata.rhui_info.target_client_pkg_names)

        def is_repofile(path):
            return os.path.dirname(path) == '/etc/yum.repos.d' and os.path.basename(path).endswith('.repo')

        def extract_repoid_from_line(line):
            return line.split(':', 1)[1].strip()

        target_ver = api.current_actor().configuration.version.target
        setup_tasks = indata.rhui_info.target_client_setup_info.preinstall_tasks.files_to_copy_into_overlay

        yum_repos_d = context.full_path('/etc/yum.repos.d')
        all_repofiles = {os.path.join(yum_repos_d, path) for path in os.listdir(yum_repos_d) if path.endswith('.repo')}
        client_repofiles = {context.full_path(path) for path in files_provided_by_clients if is_repofile(path)}

        # Exclude repofiles used to setup the target rhui access as on some platforms the repos provided by
        # the client are not sufficient to install the client into target userspace (GCP)
        rhui_setup_repofile_tasks = [task for task in setup_tasks if task.src.endswith('repo')]
        rhui_setup_repofiles = (
            get_copy_location_from_copy_in_task(context, copy_task) for copy_task in rhui_setup_repofile_tasks
        )
        rhui_setup_repofiles = {context.full_path(repofile) for repofile in rhui_setup_repofiles}

        foreign_repofiles = all_repofiles - client_repofiles - rhui_setup_repofiles

        # Rename non-client repofiles so they will not be recognized when running dnf repolist
        for foreign_repofile in foreign_repofiles:
            os.rename(foreign_repofile, '{0}.back'.format(foreign_repofile))

        try:
            dnf_cmd = ['dnf', 'repolist', '--releasever', target_ver, '-v']
            repolist_result = context.call(dnf_cmd)['stdout']
            repoid_lines = [line for line in repolist_result.split('\n') if line.startswith('Repo-id')]
            rhui_repoids = {extract_repoid_from_line(line) for line in repoid_lines}
            rh_repoids.update(rhui_repoids)

        except CalledProcessError as err:
            details = {'err': err.stderr, 'details': str(err)}
            raise StopActorExecutionError(
                message='Failed to retrieve repoids provided by target RHUI clients.',
                details=details
            )

        finally:
            # Revert the renaming of non-client repofiles
            for foreign_repofile in foreign_repofiles:
                os.rename('{0}.back'.format(foreign_repofile), foreign_repofile)

    return rh_repoids


def gather_target_repositories(context, indata):
    """
    Get available required target repositories and inhibit or raise error if basic checks do not pass.

    In case of repositories provided by Red Hat, it's checked whether the basic
    required repositories are available (or at least defined) in the given
    context. If not, raise StopActorExecutionError.

    For the custom target repositories we expect all of them have to be defined.
    If any custom target repository is missing, raise StopActorExecutionError.

    If any repository is defined multiple times, produce the inhibitor Report
    msg.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: List of target system repoids
    :rtype: List(string)
    """
    rh_available_repoids = _get_rh_available_repoids(context, indata)
    all_available_repoids = _get_all_available_repoids(context)

    target_repoids = []
    missing_custom_repoids = []
    for target_repo in api.consume(TargetRepositories):
        for rhel_repo in target_repo.rhel_repos:
            if rhel_repo.repoid in rh_available_repoids:
                target_repoids.append(rhel_repo.repoid)
            else:
                # TODO: We shall report that the RHEL repos that we deem necessary for
                # the upgrade are not available; but currently it would just print bunch of
                # data every time as we maps EUS and other repositories as well. But these
                # do not have to be necessary available on the target system in the time
                # of the upgrade. Let's skip it for now until it's clear how we will deal
                # with it.
                pass
        for custom_repo in target_repo.custom_repos:
            if custom_repo.repoid in all_available_repoids:
                target_repoids.append(custom_repo.repoid)
            else:
                missing_custom_repoids.append(custom_repo.repoid)
    api.current_logger().debug("Gathered target repositories: {}".format(', '.join(target_repoids)))
    if not target_repoids:
        target_major_version = get_target_major_version()
        reporting.create_report([
            reporting.Title('There are no enabled target repositories'),
            reporting.Summary(
                'This can happen when a system is not correctly registered with the subscription manager'
                ' or, when the leapp --no-rhsm option has been used, no custom repositories have been'
                ' passed on the command line.'
            ),
            reporting.Groups([reporting.Groups.REPOSITORY]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Remediation(hint=(
                'Ensure the system is correctly registered with the subscription manager and that'
                ' the current subscription is entitled to install the requested target version {version}.'
                ' If you used the --no-rhsm option (or the LEAPP_NO_RHSM=1 environment variable is set),'
                ' ensure the custom repository file is provided with'
                ' properly defined repositories and that the --enablerepo option for leapp is set if the'
                ' repositories are defined in any repofiles under the /etc/yum.repos.d/ directory.'
                ' For more information on custom repository files, see the documentation.'
                ' Finally, verify that the "/etc/leapp/files/repomap.json" file is up-to-date.'
            ).format(version=api.current_actor().configuration.version.target)),
            reporting.ExternalLink(
                # https://red.ht/preparing-for-upgrade-to-rhel8
                # https://red.ht/preparing-for-upgrade-to-rhel9
                # https://red.ht/preparing-for-upgrade-to-rhel10
                url='https://red.ht/preparing-for-upgrade-to-rhel{}'.format(target_major_version),
                title='Preparing for the upgrade'),
            reporting.RelatedResource("file", "/etc/leapp/files/repomap.json"),
            reporting.RelatedResource("file", "/etc/yum.repos.d/")
        ])
        raise StopActorExecution()
    if missing_custom_repoids:
        reporting.create_report([
            reporting.Title('Some required custom target repositories have not been found'),
            reporting.Summary(
                'This can happen when a repository ID was entered incorrectly either'
                ' while using the --enablerepo option of leapp, or in a third party actor that produces a'
                ' CustomTargetRepositoryMessage.\n'
                'The following repositories IDs could not be found in the target configuration:\n'
                '- {}\n'.format('\n- '.join(missing_custom_repoids))
            ),
            reporting.Groups([reporting.Groups.REPOSITORY]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.ExternalLink(
                # NOTE: Article covers both RHEL 7 to RHEL 8 and RHEL 8 to RHEL 9
                url='https://access.redhat.com/articles/4977891',
                title='Customizing your Red Hat Enterprise Linux in-place upgrade'),
            reporting.Remediation(hint=(
                'Consider using the custom repository file, which is documented in the official'
                ' upgrade documentation. Check whether a repository ID has been'
                ' entered incorrectly with the --enablerepo option of leapp.'
                ' Check the leapp logs to see the list of all available repositories.'
            ))
        ])
        raise StopActorExecution()
    return set(target_repoids)


def _install_custom_repofiles(context, custom_repofiles):
    """
    Install the required custom repository files into the container.

    The repository files are copied from the host into the /etc/yum.repos.d
    directory into the container.

    :param context: the container where the repofiles should be copied
    :type context: mounting.IsolatedActions class
    :param custom_repofiles: list of custom repo files
    :type custom_repofiles: List(CustomTargetRepositoryFile)
    """
    for rfile in custom_repofiles:
        _dst_path = os.path.join('/etc/yum.repos.d', os.path.basename(rfile.file))
        context.copy_to(rfile.file, _dst_path)


def _gather_target_repositories(context, indata, prod_cert_path):
    """
    This is wrapper function to gather the target repoids.

    Probably the function could be partially merged into gather_target_repositories
    and this could be really just wrapper with the switch of certificates.
    I am keeping that for now as it is as interim step.

    :param context: the container where the repofiles should be copied
    :type context: mounting.IsolatedActions class
    :param indata: majority of input data for the actor
    :type indata: class _InputData
    :param prod_cert_path: path where the target product cert is stored
    :type prod_cert_path: string
    """
    rhsm.set_container_mode(context)
    rhsm.switch_certificate(context, indata.rhsm_info, prod_cert_path)

    _install_custom_repofiles(context, indata.custom_repofiles)
    return gather_target_repositories(context, indata)


def _copy_files(context, files):
    """
    Copy the files/dirs from the host to the `context` userspace

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param files: list of files that should be copied from the host to the context
    :type files: list of CopyFile
    """
    for file_task in files:
        if not file_task.dst:
            file_task.dst = file_task.src
        if os.path.isdir(file_task.src):
            context.remove_tree(file_task.dst)
            context.copytree_to(file_task.src, file_task.dst)
        else:
            context.copy_to(file_task.src, file_task.dst)


def _get_target_userspace():
    return constants.TARGET_USERSPACE.format(get_target_major_version())


def _create_target_userspace(context, packages, files, target_repoids):
    """Create the target userspace."""
    target_path = _get_target_userspace()
    prepare_target_userspace(context, target_path, target_repoids, list(packages))
    _prep_repository_access(context, target_path)

    with mounting.NspawnActions(base_dir=target_path) as target_context:
        _copy_files(target_context, files)
    dnfplugin.install(_get_target_userspace())

    # and do not forget to set the rhsm into the container mode again
    with mounting.NspawnActions(_get_target_userspace()) as target_context:
        rhsm.set_container_mode(target_context)


def install_target_rhui_client_if_needed(context, indata):
    if not indata.rhui_info:
        return

    target_major_version = get_target_major_version()
    userspace_dir = _get_target_userspace()
    _create_target_userspace_directories(userspace_dir)

    setup_info = indata.rhui_info.target_client_setup_info
    if setup_info.preinstall_tasks:
        preinstall_tasks = setup_info.preinstall_tasks

        for file_to_remove in preinstall_tasks.files_to_remove:
            context.remove(file_to_remove)

        for copy_info in preinstall_tasks.files_to_copy_into_overlay:
            context.makedirs(os.path.dirname(copy_info.dst), exists_ok=True)
            context.copy_to(copy_info.src, copy_info.dst)

    cmd = ['dnf', '-y']

    if setup_info.enable_only_repoids_in_copied_files and setup_info.preinstall_tasks:
        copy_tasks = setup_info.preinstall_tasks.files_to_copy_into_overlay
        copied_repofiles = [copy.src for copy in copy_tasks if copy.src.endswith('.repo')]
        copied_repoids = set()
        for repofile in copied_repofiles:
            repofile_contents = repofileutils.parse_repofile(repofile)
            copied_repoids.update(entry.repoid for entry in repofile_contents.data)

        cmd += ['--disablerepo', '*']
        for copied_repoid in copied_repoids:
            cmd.extend(('--enablerepo', copied_repoid))

    src_client_remove_steps = ['remove {0}'.format(client) for client in indata.rhui_info.src_client_pkg_names]
    target_client_install_steps = ['install {0}'.format(client) for client in indata.rhui_info.target_client_pkg_names]

    dnf_transaction_steps = src_client_remove_steps + target_client_install_steps + ['transaction run']

    cmd += [
        '--setopt=module_platform_id=platform:el{}'.format(target_major_version),
        '--setopt=keepcache=1',
        '--releasever', api.current_actor().configuration.version.target,
        '--disableplugin', 'subscription-manager',
        'shell'
    ]

    context.call(cmd, callback_raw=utils.logging_handler, stdin='\n'.join(dnf_transaction_steps))

    if setup_info.postinstall_tasks:
        for copy_info in setup_info.postinstall_tasks.files_to_copy:
            context.makedirs(os.path.dirname(copy_info.dst), exists_ok=True)
            context.call(['cp', copy_info.src, copy_info.dst])

    # Do a cleanup so there are not duplicit repoids
    files_owned_by_clients = _query_rpm_for_pkg_files(context, indata.rhui_info.target_client_pkg_names)

    for copy_task in setup_info.preinstall_tasks.files_to_copy_into_overlay:
        dest = get_copy_location_from_copy_in_task(context, copy_task)
        can_be_cleaned_up = copy_task.src not in setup_info.files_supporting_client_operation
        if dest not in files_owned_by_clients and can_be_cleaned_up:
            context.remove(dest)


@suppress_deprecation(TMPTargetRepositoriesFacts)
def perform():
    # NOTE: this one action is out of unit-tests completely; we do not use
    # in unit tests the LEAPP_DEVEL_SKIP_RHSM envar anymore
    _check_deprecated_rhsm_skip()

    indata = _InputData()
    prod_cert_path = _get_product_certificate_path()
    reserve_space = overlaygen.get_recommended_leapp_free_space(_get_target_userspace())
    with overlaygen.create_source_overlay(
            mounts_dir=constants.MOUNTS_DIR,
            scratch_dir=constants.SCRATCH_DIR,
            storage_info=indata.storage_info,
            xfs_info=indata.xfs_info,
            scratch_reserve=reserve_space) as overlay:
        with overlay.nspawn() as context:
            # Mount the ISO into the scratch container
            target_iso = next(api.consume(TargetOSInstallationImage), None)
            with mounting.mount_upgrade_iso_to_root_dir(overlay.target, target_iso):

                install_target_rhui_client_if_needed(context, indata)

                target_repoids = _gather_target_repositories(context, indata, prod_cert_path)
                _create_target_userspace(context, indata.packages, indata.files, target_repoids)
                # TODO: this is tmp solution as proper one needs significant refactoring
                target_repo_facts = repofileutils.get_parsed_repofiles(context)
                api.produce(TMPTargetRepositoriesFacts(repositories=target_repo_facts))
                # ## TODO ends here
                api.produce(UsedTargetRepositories(
                    repos=[UsedTargetRepository(repoid=repo) for repo in target_repoids]))
                api.produce(TargetUserSpaceInfo(
                    path=_get_target_userspace(),
                    scratch=constants.SCRATCH_DIR,
                    mounts=constants.MOUNTS_DIR))



import functools
import os
import random

from leapp import models, reporting
from leapp.libraries.actor import vdoconversionscanner
from leapp.libraries.common.testutils import create_report_mocked


def aslist(f):
    """ Decorator used to convert generator to list """
    @functools.wraps(f)
    def inner(*args, **kwargs):
        return list(f(*args, **kwargs))
    return inner


def _lsblk_entry(prefix, number, types, size='128G', bsize=2 ** 37):
    lsblk_entry_name = '{0}{1}'.format(prefix, number)
    return models.LsblkEntry(
        name=lsblk_entry_name,
        kname=lsblk_entry_name,
        maj_min='253:{0}'.format(number),
        rm='0',
        size=size,
        bsize=bsize,
        ro='0',
        tp=types[random.randint(0, len(types) - 1)],
        mountpoint='')


@aslist
def _lsblk_entries(pre=0, post=0, complete=0, undetermined=0, small=0, missing=0):

    begin = pre
    for x in range(begin):
        yield _lsblk_entry('vdo_pre_', x, ['disk', 'part'])
    begin += pre

    for x in range(begin, begin + complete):
        yield _lsblk_entry('vdo_post_complete_', x, ['disk', 'part'])
    begin += complete

    for x in range(begin, begin + (post - complete)):
        yield _lsblk_entry('vdo_post_', x, ['disk', 'part'])
    begin += post - complete

    for x in range(begin, begin + undetermined):
        yield _lsblk_entry('vdo_undetermined_', x, ['disk', 'part'])
    begin += undetermined

    for x in range(begin, begin + small):
        # this one will not be undetermined. We want to test this one is actually really
        # skipped, so we do not find it in the list
        yield _lsblk_entry('vdo_undetermined_small_skipped', x, ['disk', 'part'], size='2K', bsize=2 ** 11)
    begin += small

    for x in range(begin, begin + missing):
        yield _lsblk_entry('vdo_undetermined_missing_', x, ['disk', 'part'])


def _storage_info(pre=0, post=0, complete=0, undetermined=0, small=0, missing=0):
    return models.StorageInfo(lsblk=_lsblk_entries(pre, post, complete, undetermined, small, missing))


def _check_vdo_lvm_managed(device):
    device = os.path.split(device)[-1]
    code = 2
    if device.startswith('vdo_') and ("_post_" in device) and ("_complete_" in device):
        code = 0
    return code


def _check_vdo_pre_conversion(device):
    device = os.path.split(device)[-1]
    code = 255
    if device.startswith('vdo_'):
        code = -1 if '_undetermined_' in device else 1 if "_pre_" in device else 0
    return code


def test_check_vdo_pre_conversion(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)
    monkeypatch.setattr(os.path, 'exists', lambda dummy_file: True)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': 0})
    info = vdoconversionscanner.get_info(_storage_info(pre=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (len(info.post_conversion) == 1)
    assert not info.post_conversion[0].complete
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': 1})
    info = vdoconversionscanner.get_info(_storage_info(pre=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (len(info.pre_conversion) == 1)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': 255})
    info = vdoconversionscanner.get_info(_storage_info(pre=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': -1})
    info = vdoconversionscanner.get_info(_storage_info(pre=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list) and (len(info.undetermined_conversion) == 1)


def test_check_vdo_lvm_managed(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', lambda dummy_file: True)
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': 0})
    info = vdoconversionscanner.get_info(_storage_info(post=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (len(info.post_conversion) == 1)
    assert info.post_conversion[0].complete
    assert not info.post_conversion[0].failure
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': 2})
    info = vdoconversionscanner.get_info(_storage_info(post=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (len(info.post_conversion) == 1)
    assert not info.post_conversion[0].complete
    assert not info.post_conversion[0].failure
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)

    monkeypatch.setattr(vdoconversionscanner, 'run', lambda _, checked: {'exit_code': -1})
    info = vdoconversionscanner.get_info(_storage_info(post=1))
    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (len(info.post_conversion) == 1)
    assert not info.post_conversion[0].complete
    assert info.post_conversion[0].failure
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)


def test_lvm_package_not_installed(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: False)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: False)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)

    pre = 0
    post = 0
    complete = 0
    undetermined = 5

    info = vdoconversionscanner.get_info(_storage_info(pre, post, complete, undetermined))

    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)


def test_no_vdo_devices(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)

    info = vdoconversionscanner.get_info(_storage_info())

    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list) and (not info.undetermined_conversion)


def test_vdo_devices(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', lambda dummy_file: True)
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)

    pre = 5
    post = 7
    complete = 3
    undetermined = 2

    info = vdoconversionscanner.get_info(_storage_info(pre, post, complete, undetermined))

    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (len(info.pre_conversion) == pre)
    assert isinstance(info.post_conversion, list) and (len(info.post_conversion) == post)
    assert len([x for x in info.post_conversion if x.complete]) == complete
    assert (isinstance(info.undetermined_conversion, list) and
            (len(info.undetermined_conversion) == undetermined))


def test_vdo_package_not_installed(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: False)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)

    pre = 5
    post = 7
    complete = 3
    undetermined = 2

    info = vdoconversionscanner.get_info(_storage_info(pre, post, complete, undetermined))

    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list)
    assert len(info.undetermined_conversion) == (pre + post + undetermined)


def test_missing_devices(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)
    monkeypatch.setattr(os.path, 'exists', lambda fname: '_missing_' not in fname)

    complete = 1
    undetermined = 2
    missing = 1

    info = vdoconversionscanner.get_info(_storage_info(
        complete=complete, undetermined=undetermined, missing=missing
    ))

    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (len(info.post_conversion) == complete)
    assert isinstance(info.undetermined_conversion, list)
    assert len([x for x in info.post_conversion if x.complete]) == complete
    assert len(info.undetermined_conversion) == (undetermined + missing)
    missing_detected = False
    for item in info.undetermined_conversion:
        if 'does not exist' in item.failure and '_missing_' in item.name:
            missing_detected = True
    assert missing_detected


def test_small_partition_skip(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(vdoconversionscanner, '_lvm_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_vdo_package_installed', lambda: True)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_pre_conversion', _check_vdo_pre_conversion)
    monkeypatch.setattr(vdoconversionscanner, '_check_vdo_lvm_managed', _check_vdo_lvm_managed)
    monkeypatch.setattr(os.path, 'exists', lambda fname: '_missing_' not in fname)

    undetermined = 2
    small = 2

    info = vdoconversionscanner.get_info(_storage_info(undetermined=undetermined, small=small))

    assert isinstance(info, models.VdoConversionInfo)
    assert isinstance(info.pre_conversion, list) and (not info.pre_conversion)
    assert isinstance(info.post_conversion, list) and (not info.post_conversion)
    assert isinstance(info.undetermined_conversion, list)
    assert len(info.undetermined_conversion) == undetermined
    for item in info.undetermined_conversion:
        assert 'small' not in item.name



import re

from leapp.libraries.common.spamassassinutils import SPAMC_CONFIG_FILE
from leapp.libraries.stdlib import api


def _rewrite_spamc_config(content):
    res_lines = []
    read_ssl_option = False
    for line in content.split('\n'):
        if line.startswith('#'):
            res_lines.append(line)
            continue
        if read_ssl_option:
            line = re.sub(r'^\s*(sslv3|tlsv1)', '', line)
            read_ssl_option = False
        if re.search(r'(?<!\S)--ssl\s*$', line):
            read_ssl_option = True
        line = re.sub(r'(?<!\S)--ssl(\s+|=)(sslv3|tlsv1)(?!\S)', '--ssl', line)
        res_lines.append(line)
    res = '\n'.join(res_lines)
    return res


def migrate_spamc_config(facts, fileops, backup_func):
    """
    Removes arguments given to the --ssl option in the spamc config file.
    The file is backed up beforehand.
    """
    if facts.spamc_ssl_argument is None:
        api.current_logger().info('There is nothing to migrate in the spamc configuration file')
        return
    try:
        backup_path = backup_func(SPAMC_CONFIG_FILE)
        api.current_logger().info('spamc configuration file backup created at %s.'
                                  % backup_path)
    except (OSError, IOError) as e:
        api.current_logger().warning(
            'spamc configuration file migration will not be performed. Failed to create backup: %s' % e)
        return
    try:
        content = fileops.read(SPAMC_CONFIG_FILE)
        new_content = _rewrite_spamc_config(content)
        fileops.write(SPAMC_CONFIG_FILE, new_content)
    except (OSError, IOError) as e:
        api.current_logger().warning('Failed to migrate spamc configuration file: %s' % e)



from leapp.actors import Actor
from leapp.libraries.actor import scansystemdsource
from leapp.models import SystemdBrokenSymlinksSource, SystemdServicesInfoSource, SystemdServicesPresetInfoSource
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanSystemdSource(Actor):
    """
    Provides info about systemd on the source system

    The provided info includes information about:
    - vendor presets of services
    - systemd service files, including their state
    - broken systemd symlinks

    There is an analogous actor :class:`ScanSystemdTarget` for target system.
    """

    name = 'scan_systemd_source'
    consumes = ()
    produces = (SystemdBrokenSymlinksSource, SystemdServicesInfoSource, SystemdServicesPresetInfoSource)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scansystemdsource.scan()



from leapp.libraries.actor.opensslconfigcheck import (
    _find_duplicates,
    _find_pair,
    _key_equal,
    _normalize_key,
    _openssl_duplicate_keys_in,
    _openssl_find_block,
    _openssl_reachable_block,
    _openssl_reachable_block_root,
    _openssl_reachable_key,
    _openssl_reachable_path
)
from leapp.models import OpenSslConfig, OpenSslConfigBlock, OpenSslConfigPair


def _setup_config():
    return OpenSslConfig(
        openssl_conf="default_modules",
        blocks=[
            OpenSslConfigBlock(
                name="default_modules",
                pairs=[
                    OpenSslConfigPair(
                        key="ssl_conf",
                        value="ssl_module",
                    )
                ]
            ),
            OpenSslConfigBlock(
                name="ssl_module",
                pairs=[
                    OpenSslConfigPair(
                        key="system_default",
                        value="crypto_policy",
                    )
                ]
            ),
            OpenSslConfigBlock(
                name="crypto_policy",
                pairs=[
                    OpenSslConfigPair(
                        key=".include",
                        value="/etc/crypto-policies/back-ends/opensslcnf.config",
                    )
                ]
            ),
            OpenSslConfigBlock(
                name="tsa",
                pairs=[
                    OpenSslConfigPair(
                        key="1.default_tsa",
                        value="tsa_config1",
                    ),
                    OpenSslConfigPair(
                        key="2.default_tsa",
                        value="tsa_config2",
                    )
                ]
            ),
            OpenSslConfigBlock(
                name="ca",
                pairs=[
                    OpenSslConfigPair(
                        key="default_ca",
                        value="CA_default",
                    ),
                ]
            ),
            OpenSslConfigBlock(
                name="CA_default",
                pairs=[
                    OpenSslConfigPair(
                        key="x509_extensions",
                        value="usr_cert",
                    ),
                ]
            ),
            OpenSslConfigBlock(
                name="usr_cert",
                pairs=[
                    OpenSslConfigPair(
                        key="subjectKeyIdentifier",
                        value="hash",
                    ),
                    OpenSslConfigPair(
                        key="basicConstraints",
                        value="critical,CA:true",
                    ),
                    OpenSslConfigPair(
                        key="keyUsage",
                        value="nonRepudiation, digitalSignature, keyEncipherment",
                    ),
                    OpenSslConfigPair(
                        key="subjectAltName",
                        value="email:copy",
                    ),
                    OpenSslConfigPair(
                        key="subjectAltName",
                        value="email:move",
                    ),  # yay, duplicate
                ]
            ),
            OpenSslConfigBlock(
                name="req",
                pairs=[
                    OpenSslConfigPair(
                        key="default_bits",
                        value="2048",
                    ),
                    OpenSslConfigPair(
                        key="default_md",
                        value="sha256",
                    ),
                    OpenSslConfigPair(
                        key="x509_extensions",
                        value="v3_ca",
                    ),
                ]
            ),
            OpenSslConfigBlock(
                name="v3_ca",
                pairs=[
                    OpenSslConfigPair(
                        key="subjectKeyIdentifier",
                        value="hash",
                    ),
                    OpenSslConfigPair(
                        key="basicConstraints",
                        value="critical,CA:true",
                    ),
                    OpenSslConfigPair(
                        key="keyUsage",
                        value="cRLSign, keyCertSign",
                    ),
                    OpenSslConfigPair(
                        key="keyUsage",
                        value="cRLSign",
                    ),  # yay, duplicate
                    OpenSslConfigPair(
                        key="keyUsage",
                        value="keyCertSign",
                    ),  # yay, duplicate
                ]
            ),
        ]
    )


def test_normalize_key():
    assert _normalize_key("default_tsa") == "default_tsa"

    # TODO this is questionable, but probably ok
    assert _normalize_key(".include") == "include"

    assert _normalize_key("DTLS.MaxProtocol") == "MaxProtocol"

    assert _normalize_key("DTLS.TLS.MaxProtocol") == "TLS.MaxProtocol"


def test_key_equal():
    pair = OpenSslConfigPair(key="default_tsa", value="tsa_config1")
    assert _key_equal(pair, "default_tsa")

    pair = OpenSslConfigPair(key="default_tsa", value="tsa_config1")
    assert not _key_equal(pair, "tsa")

    pair = OpenSslConfigPair(key=".include", value="/some/path")
    assert _key_equal(pair, ".include")

    # TODO this is questionable, but probably ok
    pair = OpenSslConfigPair(key=".include", value="/some/path")
    assert _key_equal(pair, "include")

    pair = OpenSslConfigPair(key="DTLS.MaxProtocol", value="DTLSv1.2")
    assert _key_equal(pair, "MaxProtocol")

    pair = OpenSslConfigPair(key="DTLS.TLS.MaxProtocol", value="DTLSv1.2")
    assert not _key_equal(pair, "MaxProtocol")


def test_find():
    config = _setup_config()

    result = _openssl_find_block(config, 'default_modules')
    assert result.name == 'default_modules'
    pair = _find_pair(result, 'ssl_conf')
    assert pair.key == 'ssl_conf'
    assert pair.value == 'ssl_module'

    result = _openssl_find_block(config, 'ssl_module')
    assert result.name == 'ssl_module'
    pair = _find_pair(result, 'system_default')
    assert pair.key == 'system_default'
    assert pair.value == 'crypto_policy'

    result = _openssl_find_block(config, 'crypto_policy')
    assert result.name == 'crypto_policy'
    pair = _find_pair(result, '.include')
    assert pair.key == '.include'
    assert pair.value == '/etc/crypto-policies/back-ends/opensslcnf.config'

    assert _find_pair(result, 'nonexisting') is None

    assert _openssl_find_block(config, 'nonexisting') is None

    # duplicates
    b = _openssl_find_block(config, "tsa")
    assert b
    p = _find_pair(b, "default_tsa")
    assert p.key == "2.default_tsa"
    assert p.value == "tsa_config2"


def test_reachable_block():
    config = _setup_config()

    assert _openssl_reachable_block_root(config, "default_modules")

    assert _openssl_reachable_block_root(config, "ssl_module")

    assert _openssl_reachable_block_root(config, "crypto_policy")

    assert not _openssl_reachable_block_root(config, "crypto_policy", 1)

    assert not _openssl_reachable_block_root(config, "tsa")

    # searching from arbitrary block
    assert _openssl_reachable_block(config, "tsa", "tsa_config1")

    # non-existent start key
    assert not _openssl_reachable_block(config, "ecdsa", "crypto_policy")


def test_reachable_path():
    config = _setup_config()

    assert _openssl_reachable_path(config,
                                   ("default_modules", "ssl_conf", "ssl_module", "system_default",
                                    "crypto_policy", ".include"),
                                   "/etc/crypto-policies/back-ends/opensslcnf.config")

    assert _openssl_reachable_path(config,
                                   ("ssl_module", "system_default"),
                                   "crypto_policy")

    # missing key in the path
    assert not _openssl_reachable_path(config,
                                       ("ssl_module", "system_default", "crypto_policy"),
                                       "crypto_policy")

    # non continuous path/wrong value in path
    assert not _openssl_reachable_path(config,
                                       ("ssl_module", "tsa"),
                                       "tsa_config1")

    # wrong value
    assert not _openssl_reachable_path(config,
                                       ("ssl_module", "system_default"),
                                       ".include")

    # not reachable from the root
    assert not _openssl_reachable_path(config,
                                       ("tsa", "default_tsa"),
                                       "tsa_config1")

    # any value
    assert _openssl_reachable_path(config, ("ssl_module", "system_default"))
    assert not _openssl_reachable_path(config, ("ssl_module", "tsa"))


def test_reachable_key():
    config = _setup_config()

    assert _openssl_reachable_key(config, "system_default")
    assert _openssl_reachable_key(config, "system_default", "crypto_policy")
    assert not _openssl_reachable_key(config, "tsa")
    assert not _openssl_reachable_key(config, "default_tsa", "tsa_config2")


def test_find_duplicates():
    config = _setup_config()

    d = _find_duplicates(config, "non-existent")
    assert len(d) == 0

    d = _find_duplicates(config, "default_modules")
    assert len(d) == 0

    d = _find_duplicates(config, "tsa")
    assert len(d) == 1
    assert d[0] == 'default_tsa'

    d = _find_duplicates(config, "v3_ca")
    assert len(d) == 1
    assert d[0] == 'keyUsage'

    d = _find_duplicates(config, "usr_cert")
    assert len(d) == 1
    assert d[0] == 'subjectAltName'


def test_duplicate_keys_in():
    config = _setup_config()

    d = _openssl_duplicate_keys_in(config, "x509_extensions")
    assert len(d) == 2
    assert 'keyUsage' in d
    assert 'subjectAltName' in d



import pytest

from leapp.libraries.actor import enabledeviceciofreeservice
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import SystemdServicesTasks


def test_task_produced_on_s390(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch=architecture.ARCH_S390X))
    monkeypatch.setattr(api, "produce", produce_mocked())

    enabledeviceciofreeservice.process()

    assert api.produce.called
    assert isinstance(api.produce.model_instances[0], SystemdServicesTasks)
    assert api.produce.model_instances[0].to_enable == ['device_cio_free.service']


@pytest.mark.parametrize('arch', [
    architecture.ARCH_X86_64,
    architecture.ARCH_ARM64,
    architecture.ARCH_PPC64LE,
])
def test_task_not_produced_on_non_s390(monkeypatch, arch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch=arch))
    monkeypatch.setattr(api, "produce", produce_mocked())

    enabledeviceciofreeservice.process()

    assert not api.produce.called
